/******************************************************************************
 * ParsedMsg.hh
 * Copyright (c) 2003 Virtual Werder
 *
 *
 * implements:
 *   class ParsedMsg
 *   class InitSideMsg
 *
 * requires:
 *
 *
 * Created on: 07.05.2003 by pagra
 * Maintainer: pagra
 * Group:
 * Co-authors:
 *****************************************************************************/

#ifndef PARSEDMSG_HH
#define PARSEDMSG_HH

#include "common.hh"
#include <vector>
using namespace std;

class ParsedBall;
class ParsedPlayer;

/**
 ParsedMsg provides basic information of parsed message of the parser
 @ingroup Parser
 @brief class ParsedMsg in group Parser
 */
class ParsedMsg {
public:
  ParsedMsg( ParsedMsgType msgType = Unknown_Msg );
  ~ParsedMsg();

  inline void setType( ParsedMsgType msgType ) { this->msgType = msgType; }
  inline ParsedMsgType getType() const { return msgType; };
  
private:
  ParsedMsgType msgType;
};


/**
 @ingroup Parser
 @brief class InitSideOk in group Parser
 */
class InitSideOkMsg : public ParsedMsg {
public:
  InitSideOkMsg();
  ~InitSideOkMsg();

  FieldSide fieldSide;
};


/**
 @ingroup Parser
 @brief class OkEyeMsg in group Parser
 */
class OkEyeMsg : public ParsedMsg {
public:
  OkEyeMsg();
  ~OkEyeMsg();
  
  bool eyeOn;
};


/**
 @ingroup Parser
 @brief class OkChangePlayerTypeMsg in group Parser
 */
class OkChangePlayerTypeMsg : public ParsedMsg {
public:
  OkChangePlayerTypeMsg();
  ~OkChangePlayerTypeMsg();
  
  int uNum;
  int playerType;
};


/**
 @ingroup Parser
 @brief class ChangePlayerTypeMsg in group Parser
 */
class ChangePlayerTypeMsg : public ParsedMsg {
public:
  ChangePlayerTypeMsg();
  ~ChangePlayerTypeMsg();
  
  int uNum;
};


/**
 @ingroup Parser
 @brief class WNoTeamFoundMsg in group Parser
 */
class WNoTeamFoundMsg : public ParsedMsg {
public:
  WNoTeamFoundMsg();
  ~WNoTeamFoundMsg();
};



/**
 @ingroup Parser
 @brief class WNoSuchPlayerMsg in group Parser
 */
class WNoSuchPlayerMsg : public ParsedMsg {
public:
  WNoSuchPlayerMsg();
  ~WNoSuchPlayerMsg();
};



/**
 @ingroup Parser
 @brief class WCannotSubWhilePlayOnMsg in group Parser
 */
class WCannotSubWhilePlayOnMsg : public ParsedMsg {
public:
  WCannotSubWhilePlayOnMsg();
  ~WCannotSubWhilePlayOnMsg();
};


/**
 @ingroup Parser
 @brief class WNoSubsLeftMsg in group Parser
 */
class WNoSubsLeftMsg : public ParsedMsg {
public:
  WNoSubsLeftMsg();
  ~WNoSubsLeftMsg();
};


/**
 @ingroup Parser
 @brief class WMaxOfThatPlayerTypeOnFieldMsg in group Parser
 */
class WMaxOfThatPlayerTypeOnFieldMsg : public ParsedMsg {
public:
  WMaxOfThatPlayerTypeOnFieldMsg();
  ~WMaxOfThatPlayerTypeOnFieldMsg();
};

/**
 @ingroup Parser
 @brief class WCannotChangeGoalieMsg in group Parser
 */
class WCannotChangeGoalieMsg : public ParsedMsg {
public:
  WCannotChangeGoalieMsg();
  ~WCannotChangeGoalieMsg();
};

/**
 @ingroup Parser
 @brief class OkTeamnamesMsg in group Parser
 */
class OkTeamnamesMsg : public ParsedMsg {
public:
  OkTeamnamesMsg();
  ~OkTeamnamesMsg();

  void print() const;
  
  char *teamnameLeft;
  char *teamnameRight;
};


/**
 @ingroup Parser
 @brief class OkSayMsg in group Parser
 */
class OkSayMsg : public ParsedMsg {
public:
  OkSayMsg();
  ~OkSayMsg();
};

/**
 Message with parsed server parameters
 @ingroup Parser
 @brief class ServerParamMsg in group Parser
 */
class ServerParamMsg : public ParsedMsg {
public:
  ServerParamMsg();
  ~ServerParamMsg();
  
  double audio_cut_dist;
  int back_passes;
  double ball_accel_max;
  double ball_decay;
  double ball_rand;
  double ball_size;
  double ball_speed_max;
  double ball_weight;
  int catch_ban_cycle;
  double catch_probability;
  double catchable_area_l;
  double catchable_area_w;
  double ckick_margin;
  int clang_advice_win;
  int clang_define_win;
  int clang_del_win;
  int clang_info_win;
  int clang_mess_delay;
  int clang_mess_per_cycle;
  int clang_meta_win;
  int clang_rule_win;
  int clang_win_size;
  double dash_power_rate;
  int drop_ball_time;
  double effort_dec;
  double effort_dec_thr;
  double effort_inc;
  double effort_inc_thr;
  double effort_init;
  double effort_min;
  int forbid_kick_off_offside;
  int free_kick_faults;
  int freeform_send_period;
  int freeform_wait_period;
  int fullstate_l;
  int fullstate_r;
  double goal_width;
  int goalie_max_moves;
  int half_time;
  int hear_decay;
  int hear_inc;
  int hear_max;
  double inertia_moment;
  double kick_power_rate;
  double kick_rand;
  double kick_rand_factor_l;
  double kick_rand_factor_r;
  double kickable_margin;
  int max_goal_kicks;
  double maxmoment;
  double maxneckang;
  double maxneckmoment;
  double maxpower;
  double minmoment;
  double minneckang;
  double minneckmoment;
  double minpower;
  double offside_active_area_size;
  double offside_kick_margin;
  double player_accel_max;
  double player_decay;
  double player_rand;
  double player_size;
  double player_speed_max;
  double player_weight;
  int point_to_ban;
  int point_to_duration;
  double prand_factor_l;
  double prand_factor_r;
  int proper_goal_kicks;
  double quantize_step;
  double quantize_step_l;
  double recover_dec;
  double recover_dec_thr;
  double recover_min;
  double recv_step;
  int say_coach_cnt_max;
  int say_coach_msg_size;
  int say_msg_size;
  int send_comms;
  int send_step;
  int send_vi_step;
  int sense_body_step;
  int simulator_step;
  double slow_down_factor;
  int slowness_on_top_for_left_team;
  int slowness_on_top_for_right_team;
  double stamina_inc_max;
  double stamina_max;
  int start_goal_l;
  int start_goal_r;
  double stopped_ball_vel;
  int synch_micro_sleep;
  int synch_mode;
  int synch_offset;
  double tackle_back_dist;
  int tackle_cycles;
  double tackle_dist;
  double tackle_exponent;
  double tackle_power_rate;
  double tackle_width;
  int use_offside;
  double visible_angle;
  double visible_distance;
  double wind_ang;
  double wind_dir;
  double wind_force;
  int wind_none;
  double wind_rand;
  int wind_random;
};


/**
 Message containing parsed player parameters
 @ingroup Parser
 @brief class PlayerParamMsg in group Parser
 */
class PlayerParamMsg : public ParsedMsg {
public:
  PlayerParamMsg();
  ~PlayerParamMsg();
  
  void print() const;

  int player_types;
  int pt_max;
  int random_seed;
  int subs_max;
  
  double dash_power_rate_delta_max;
  double dash_power_rate_delta_min;
  double effort_max_delta_factor;
  double effort_min_delta_factor;
  double extra_stamina_delta_max;
  double extra_stamina_delta_min;
  double inertia_moment_delta_factor;
  double kick_rand_delta_factor;
  double kickable_margin_delta_max;
  double kickable_margin_delta_min;
  double new_dash_power_rate_delta_max;
  double new_dash_power_rate_delta_min;
  double new_stamina_inc_max_delta_factor;
  double player_decay_delta_max;
  double player_decay_delta_min;
  double player_size_delta_factor;
  double player_speed_max_delta_max;
  double player_speed_max_delta_min;
  double stamina_inc_max_delta_factor;
};

/**
 Message representing a player type
 @ingroup Parser
 @brief class PlayerTypeMsg in group Parser
 */
class PlayerTypeMsg : public ParsedMsg {
public:
  PlayerTypeMsg();
  ~PlayerTypeMsg();

  int id ;
  double player_speed_max ;
  double stamina_inc_max ;
  double player_decay ;
  double inertia_moment ;
  double dash_power_rate ;
  double player_size ;
  double kickable_margin ;
  double kick_rand ; 
  double extra_stamina ;
  double effort_max ;
  double effort_min ;
};


/**
 Message providing all basic information of visual message, like
 player positions, current cycle, ball
 @ingroup Parser
 @brief class VisualMsg in group Parser
 */
class VisualMsg : public ParsedMsg {
public:
  VisualMsg( ParsedMsgType msgType = Unknown_Msg );
  ~VisualMsg();

  void print() const;
  void printBallAndTeams() const;

  int serverTime;
  ParsedBall *ball;
  vector<ParsedPlayer *> *ourTeam;
  vector<ParsedPlayer *> *theirTeam;
  
};

/**
 received visual message is a see global message
 @ingroup Parser
 @brief class SeeGlobalMsg in group Parser
*/
class SeeGlobalMsg : public VisualMsg {
public:
  SeeGlobalMsg();
  ~SeeGlobalMsg();

  void print() const;
};


/**
 received visual message is a ok look message
 @ingroup Parser
 @brief class OkLookMsg in group Parser
*/
class OkLookMsg : public VisualMsg {
public: 
  OkLookMsg();
  ~OkLookMsg();

  void print() const;
};
  

/**
 @ingroup Parser
 @brief class HearRefereeMsg in group Parser
*/
class HearRefereeMsg : public ParsedMsg {
public:
  HearRefereeMsg();
  ~HearRefereeMsg();

  void print() const;

  int serverTime;
  Playmode playmode;
};


/**
 @ingroup Parser
 @brief class HearPlayerMsg in group Parser
*/
class HearPlayerMsg : public ParsedMsg {
public:
  HearPlayerMsg();
  ~HearPlayerMsg();

  void print() const;

  int serverTime;
  int uNum;
  bool ourTeam;
  char *message;
};


class TeamGraphicOkMsg : public ParsedMsg {
public:
  TeamGraphicOkMsg();
  ~TeamGraphicOkMsg();

  int x;
  int y;
};

/**
 @ingroup Parser
 @brief class CLangVersPlayerMsg in group Parser
*/
class CLangVersPlayerMsg : public ParsedMsg {
public:
  CLangVersPlayerMsg();
  ~CLangVersPlayerMsg();

  void print() const;
  
  int uNum;
  bool ourTeam;
  int maxCLangVers;
  int minCLangVers;
};


/**
 @ingroup Parser
 @brief class ENoMoreTeamOrPlayerOrGoalieMsg in group Parser
*/
class ENoMoreTeamOrPlayerOrGoalieMsg : public ParsedMsg {
public:
  ENoMoreTeamOrPlayerOrGoalieMsg();
  ~ENoMoreTeamOrPlayerOrGoalieMsg();
};


/**
 @ingroup Parser
 @brief class ENoSuchTeamOrAlreadyHaveCoachMsg in group Parser
*/
class ENoSuchTeamOrAlreadyHaveCoachMsg : public ParsedMsg {
public:
  ENoSuchTeamOrAlreadyHaveCoachMsg();
  ~ENoSuchTeamOrAlreadyHaveCoachMsg();
};


/**
 @ingroup Parser
 @brief class EUnknownCommandMsg in group Parser
*/
class EUnknownCommandMsg : public ParsedMsg {
public:
  EUnknownCommandMsg();
  ~EUnknownCommandMsg();
};


/**
 @ingroup Parser
 @brief class EIllegalCommandFormMsg in group Parser
*/
class EIllegalCommandFormMsg : public ParsedMsg {
public:
  EIllegalCommandFormMsg();
  ~EIllegalCommandFormMsg();
};


/**
 @ingroup Parser
 @brief class EOutOfRangePlayerTypeMsg in group Parser
*/
class EOutOfRangePlayerTypeMsg : public ParsedMsg {
public:
  EOutOfRangePlayerTypeMsg();
  ~EOutOfRangePlayerTypeMsg();
};

/**
 @ingroup Parser
 @brief class EIllegalModeMsg in group Parser
*/
class EIllegalModeMsg : public ParsedMsg {
public:
  EIllegalModeMsg();
  ~EIllegalModeMsg();
};

/**
 @ingroup Parser
 @brief class EIllegalObjectFormMsg in group Parser
*/
class EIllegalObjectFormMsg : public ParsedMsg {
public:
  EIllegalObjectFormMsg();
  ~EIllegalObjectFormMsg();
};

/**
 @ingroup Parser
 @brief class ESaidTooManyFreeformMessagesMsg in group Parser
*/
class ESaidTooManyFreeformMessagesMsg : public ParsedMsg {
public:
  ESaidTooManyFreeformMessagesMsg();
  ~ESaidTooManyFreeformMessagesMsg();
};


/**
 @ingroup Parser
 @brief class ECannotSayFreeformWhilePlayonMsg in group Parser
*/
class ECannotSayFreeformWhilePlayonMsg : public ParsedMsg {
public:
  ECannotSayFreeformWhilePlayonMsg();
  ~ECannotSayFreeformWhilePlayonMsg();
};


/**
 @ingroup Parser
 @brief class ESaidTooManyMetaMessagesMsg in group Parser
*/
class ESaidTooManyMetaMessagesMsg : public ParsedMsg {
public:
  ESaidTooManyMetaMessagesMsg();
  ~ESaidTooManyMetaMessagesMsg();
};


/**
 @ingroup Parser
 @brief class ESaidTooManyAdviceMessagesMsg in group Parser
*/
class ESaidTooManyAdviceMessagesMsg : public ParsedMsg {
public:
  ESaidTooManyAdviceMessagesMsg();
  ~ESaidTooManyAdviceMessagesMsg();
};


/**
 @ingroup Parser
 @brief class ESaidTooManyInfoMessagesMsg in group Parser
*/
class ESaidTooManyInfoMessagesMsg : public ParsedMsg {
public:
  ESaidTooManyInfoMessagesMsg();
  ~ESaidTooManyInfoMessagesMsg();
};


/**
 @ingroup Parser
 @brief class ESaidTooManyDefineMessagesMsg in group Parser
*/
class ESaidTooManyDefineMessagesMsg : public ParsedMsg {
public:
  ESaidTooManyDefineMessagesMsg();
  ~ESaidTooManyDefineMessagesMsg();
};


/**
 @ingroup Parser
 @brief class ESaidTooManyDelMessagesMsg in group Parser
*/
class ESaidTooManyDelMessagesMsg : public ParsedMsg {
public:
  ESaidTooManyDelMessagesMsg();
  ~ESaidTooManyDelMessagesMsg();
};


/**
 @ingroup Parser
 @brief class ESaidTooManyRuleMessagesMsg in group Parser
*/
class ESaidTooManyRuleMessagesMsg : public ParsedMsg {
public:
  ESaidTooManyRuleMessagesMsg();
  ~ESaidTooManyRuleMessagesMsg();
};


/**
 @ingroup Parser
 @brief class ECouldNotParseSayMsg in group Parser
*/
class ECouldNotParseSayMsg : public ParsedMsg {
public:
  ECouldNotParseSayMsg();
  ~ECouldNotParseSayMsg();
};


/**
 @ingroup Parser
 @brief class ESaidTooManyMessagesMsg in group Parser
*/
class ESaidTooManyMessagesMsg : public ParsedMsg {
public:
  ESaidTooManyMessagesMsg();
  ~ESaidTooManyMessagesMsg();
};


/**
 @ingroup Parser
 @brief class EUnknownErrorMsg in group Parser
*/
class EUnknownErrorMsg : public ParsedMsg {
public:
  EUnknownErrorMsg( const char* msg );
  ~EUnknownErrorMsg();

  char *errorMsg;
};

#endif // PARSEDMSG_HH
