/******************************************************************************
 * Parser.cc
 * Copyright (c) 2003 Virtual Werder
 *
 *
 * implements:
 *   class Parser
 *
 * requires:
 *
 *
 * Created on: 09.05.2003 by pagra
 * Maintainer: pagra
 * Group:
 * Co-authors: josh
 *****************************************************************************/


#include "Parser.hh"
#include "Communicator.hh"
#include "debug.hh"
#include "ConfigFile.hh"
#include "ParsedMsg.hh"
#include "ParsedObjects.hh"
#include <string.h>


Parser::Parser() {
  ourSide = FS_Unknown;
}

Parser::~Parser() {}


ParsedMsg *Parser::parseMsg( char *msg ) {

  if ( strncmp (msg, "(ok team_graphic ", 17) == 0)
    return parseTeamGraphicOkMsg( msg );

  if ( strncmp( msg, "(hear ", 6 ) == 0 )
    return parseHearMsg( msg );
  
  if ( strncmp( msg, "(see_g", 6 ) == 0 )
    return parseSeeGlobalMsg( msg + 12 );

  if ( strncmp( msg, "(ok say)", 8 ) == 0 )
    return new OkSayMsg();

  if ( strncmp( msg, "(ok look ", 9 ) == 0 )
    return parseOkLookMsg( msg + 9 );
  
  if ( strncmp( msg, "(ok team_names ", 15 ) == 0 )
    return parseOkTeamnamesMsg( msg + 15 );

  if ( strncmp( msg, "(clang (ver (p ", 15 ) == 0 )
    return parseCLangVersPlayerMsg( msg + 16 );

  if ( strncmp( msg, "(ok change_player_type ", 23 ) == 0 )
    return parseOkChangePlayerTypeMsg( msg + 23 );
  
  if ( strncmp( msg, "(change_player_type ", 20 ) == 0 )
    return parseChangePlayerTypeMsg( msg + 20 );
  
  if ( strncmp( msg, "(warning no_team_found)", 23 ) == 0 )
    return new WNoTeamFoundMsg();
  if ( strncmp( msg, "(warning no_such_player)", 24 ) == 0 )
    return new WNoSuchPlayerMsg();
  if ( strncmp( msg, "(warning cannot_sub_while_playon)", 33 ) == 0 )
    return new WCannotSubWhilePlayOnMsg();
  if ( strncmp( msg, "(warning no_subs_left)", 22 ) == 0 )
    return new WNoSubsLeftMsg();
  if ( strncmp( msg, "(warning max_of_that_type_on_field)", 35 ) == 0 )
    return new WMaxOfThatPlayerTypeOnFieldMsg();
  if ( strncmp( msg, "(warning cannot_change_goalie)", 30 ) == 0 )
    return new WCannotChangeGoalieMsg();
  
  if ( strncmp( msg, "(player_type ", 13 ) == 0 ) {
    return parsePlayerTypeMsg( msg + 13 );
  }

  if ( strncmp( msg, "(player_param ", 14 ) == 0 ) {
    return parsePlayerParamMsg( msg + 14 );
  }

  if ( strncmp( msg, "(server_param ", 14 ) == 0 ) {
    return parseServerParamMsg( msg + 14 );
  }

  if ( strncmp( msg, "(think)", 7 ) == 0 ){
    ParsedMsg *parsedMsg = new ParsedMsg( Think_Msg );
    return parsedMsg;
  }
  
  if ( strncmp( msg, "(init l ok)", 11 ) == 0 ) {
    InitSideOkMsg *parsedMsg = new InitSideOkMsg();
    parsedMsg->fieldSide = FS_Left;
    ourSide = FS_Left;
    return parsedMsg;
  }
  if ( strncmp( msg, "(init r ok)", 11 ) == 0 ) {
    InitSideOkMsg *parsedMsg = new InitSideOkMsg();
    parsedMsg->fieldSide = FS_Right;
    ourSide = FS_Right;
    return parsedMsg;
  }

  if ( strncmp( msg, "(ok eye on)", 11 ) == 0 ) {
    OkEyeMsg *parsedMsg = new OkEyeMsg();
    parsedMsg->eyeOn = true;
    return parsedMsg;
  }
  if ( strncmp( msg, "(ok eye off)", 12 ) == 0 ) {
    OkEyeMsg *parsedMsg = new OkEyeMsg();
    parsedMsg->eyeOn = false;
    return parsedMsg;
  }

  if ( strncmp( msg, "(error no_more_team_or_player_or_goalie)", 40 ) == 0 ) 
    return new ENoMoreTeamOrPlayerOrGoalieMsg();
  if ( strncmp( msg, "(error no_such_team_or_already_have_coach)", 42) == 0 ) 
    return new ENoSuchTeamOrAlreadyHaveCoachMsg();
  if ( strncmp( msg, "(error unknown_command)", 23) == 0 )
    return new EUnknownCommandMsg();
  if ( strncmp( msg, "(error illegal_command_form)", 28 ) == 0 )
    return new EIllegalCommandFormMsg();
  if ( strncmp( msg, "(error out_of_range_player_type)", 32) == 0 ) 
    return new EOutOfRangePlayerTypeMsg();
  if ( strncmp( msg, "(error illegal_mode)", 20 ) == 0 )
    return new EIllegalModeMsg();
  if ( strncmp( msg, "(error illegal_object_form)", 27) == 0 ) 
    return new EIllegalObjectFormMsg();
  if ( strncmp( msg, "(error said_too_many_freeform_messages)", 39) == 0 ) 
    return new ESaidTooManyFreeformMessagesMsg();
  if ( strncmp( msg, "(error cannot_say_freeform_while_playon)", 40) == 0 ) 
    return new ECannotSayFreeformWhilePlayonMsg();
  if ( strncmp( msg, "(error said_too_many_meta_messages)", 35) == 0 ) 
    return new ESaidTooManyMetaMessagesMsg();
  if ( strncmp( msg, "(error said_too_many_advice_messages)", 37) == 0 ) 
    return new ESaidTooManyAdviceMessagesMsg();
  if ( strncmp( msg, "(error said_too_many_info_messages)", 35) == 0 ) 
    return new ESaidTooManyInfoMessagesMsg();
  if ( strncmp( msg, "(error said_too_many_define_messages)", 37) == 0 ) 
    return new ESaidTooManyDefineMessagesMsg();
  if ( strncmp( msg, "(error said_too_many_del_messages)", 34) == 0 ) 
    return new ESaidTooManyDelMessagesMsg();
  if ( strncmp( msg, "(error said_too_many_rule_messages)", 35) == 0 ) 
    return new ESaidTooManyRuleMessagesMsg();
  if ( strncmp( msg, "(error could_not_parse_say)", 27) == 0 ) 
    return new ECouldNotParseSayMsg();
  if ( strncmp( msg, "(error said_too_many_messages)", 30) == 0 ) 
    return new ESaidTooManyMessagesMsg();
  if ( strncmp( msg, "(error ", 7) == 0 )
    return new EUnknownErrorMsg( msg );

  DEBUG2("Can't read message: ", msg);
  // default: can't parse message
  ParsedMsg *parsedMsg = new ParsedMsg( Unknown_Msg );
  return parsedMsg;
}


/* Parses these messages:
 * (hear servertime referee playmode)
 * (hear servertime (p "teamname" uNum) "saymsg")
 */
ParsedMsg *Parser::parseHearMsg( char *msg ) const {
  
  char *posReferee = strstr( msg, "referee" );


  // hear referee msg -> playmode changed 
  // (for playmode overview see types.h in soccerserver and our common.hh)
  if ( posReferee != NULL ) {
    HearRefereeMsg *parsedMsg = new HearRefereeMsg();
    sscanf( msg, "(hear %d", &(parsedMsg->serverTime) );
    char *posPlaymode = posReferee + 8;
    if ( strncmp( posPlaymode, "before_kick_off", 15) == 0 ) 
      parsedMsg->playmode = PM_BeforeKickOff;
    else if ( strncmp( posPlaymode, "time_over", 9) == 0 ) 
      parsedMsg->playmode = PM_TimeOver;
    else if ( strncmp( posPlaymode, "play_on", 7) == 0 ) 
      parsedMsg->playmode = PM_PlayOn;
    else if ( strncmp( posPlaymode, "kick_off_l", 10) == 0 ) {
      if ( ourSide == FS_Left )	parsedMsg->playmode = PM_OurKickOff;
      else parsedMsg->playmode = PM_TheirKickOff;
    }
    else if ( strncmp( posPlaymode, "kick_off_r", 10) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirKickOff;
      else parsedMsg->playmode = PM_OurKickOff;
    }
    else if ( strncmp( posPlaymode, "kick_in_l", 9) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurKickIn;
      else parsedMsg->playmode = PM_TheirKickIn;
    }
    else if ( strncmp( posPlaymode, "kick_in_r", 9) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirKickIn;
      else parsedMsg->playmode = PM_OurKickIn;
    }
    else if ( strncmp( posPlaymode, "free_kick_l", 11) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurFreeKick;
      else parsedMsg->playmode = PM_TheirFreeKick;
    }
    else if ( strncmp( posPlaymode, "free_kick_r", 11) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirFreeKick;
      else parsedMsg->playmode = PM_OurFreeKick;
    }
    else if ( strncmp( posPlaymode, "corner_kick_l", 12) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurCornerKick;
      else parsedMsg->playmode = PM_TheirCornerKick;
    }
    else if ( strncmp( posPlaymode, "corner_kick_r", 12) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirCornerKick;
      else parsedMsg->playmode = PM_OurCornerKick;
    }
    else if ( strncmp( posPlaymode, "goal_kick_l", 11) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurGoalKick;
      else parsedMsg->playmode = PM_TheirGoalKick;
    }
    else if ( strncmp( posPlaymode, "goal_kick_r", 11) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirGoalKick;
      else parsedMsg->playmode = PM_OurGoalKick;
    }
    else if ( strncmp( posPlaymode, "goal_l", 6) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurGoal;
      else parsedMsg->playmode = PM_TheirGoal;
    }
    else if ( strncmp( posPlaymode, "goal_r", 6) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirGoal;
      else parsedMsg->playmode = PM_OurGoal;
    }
    else if ( strncmp( posPlaymode, "drop_ball", 9) == 0 ) 
      parsedMsg->playmode = PM_DropBall;
    else if ( strncmp( posPlaymode, "offside_l", 9) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurOffside;
      else parsedMsg->playmode = PM_TheirOffside;
    }
    else if ( strncmp( posPlaymode, "offside_r", 9) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirOffside;
      else parsedMsg->playmode = PM_OurOffside;
    }
    else if ( strncmp( posPlaymode, "penalty_kick_l", 14) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurPenaltyKick;
      else parsedMsg->playmode = PM_TheirPenaltyKick;
    }
    else if ( strncmp( posPlaymode, "penalty_kick_r", 14) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirPenaltyKick;
      else parsedMsg->playmode = PM_OurPenaltyKick;
    }
    else if ( strncmp( posPlaymode, "first_half_over", 15) == 0 ) 
      parsedMsg->playmode = PM_HalfTime;
    else if ( strncmp( posPlaymode, "pause", 5) == 0 ) 
      parsedMsg->playmode = PM_Pause;
    else if ( strncmp( posPlaymode, "human_judge", 11) == 0 ) 
      parsedMsg->playmode = PM_HumanJudge;
    else if ( strncmp( posPlaymode, "foul_charge_l", 13) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurFoulCharge;
      else parsedMsg->playmode = PM_TheirFoulCharge;
    }
    else if ( strncmp( posPlaymode, "foul_charge_r", 13) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirFoulCharge;
      else parsedMsg->playmode = PM_OurFoulCharge;
    }
    else if ( strncmp( posPlaymode, "foul_push_l", 11) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurFoulPush;
      else parsedMsg->playmode = PM_TheirFoulPush;
    }
    else if ( strncmp( posPlaymode, "foul_push_r", 11) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirFoulPush;
      else parsedMsg->playmode = PM_OurFoulPush;
    }
    else if ( strncmp( posPlaymode, "foul_multiple_attack_l", 22) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurFoulMultipleAttack;
      else parsedMsg->playmode = PM_TheirFoulMultipleAttack;
    }
    else if ( strncmp( posPlaymode, "foul_multiple_attack_r", 22) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirFoulMultipleAttack;
      else parsedMsg->playmode = PM_OurFoulMultipleAttack;
    }
    else if ( strncmp( posPlaymode, "foul_ballout_l", 14) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurFoulBallout;
      else parsedMsg->playmode = PM_TheirFoulBallout;
    }
    else if ( strncmp( posPlaymode, "foul_ballout_r", 14) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirFoulBallout;
      else parsedMsg->playmode = PM_OurFoulBallout;
    }
    else if ( strncmp( posPlaymode, "back_pass_l", 11) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurBackPass;
      else parsedMsg->playmode = PM_TheirBackPass;
    }
    else if ( strncmp( posPlaymode, "back_pass_r", 11) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirBackPass;
      else parsedMsg->playmode = PM_OurBackPass;
    }
    else if ( strncmp( posPlaymode, "free_kick_fault_l", 17) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurFreeKickFault;
      else parsedMsg->playmode = PM_TheirFreeKickFault;
    }
    else if ( strncmp( posPlaymode, "free_kick_fault_r", 17) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirFreeKickFault;
      else parsedMsg->playmode = PM_OurFreeKickFault;
    }
    else if ( strncmp( posPlaymode, "catch_fault_l", 13) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurCatchFault;
      else parsedMsg->playmode = PM_TheirCatchFault;
    }
    else if ( strncmp( posPlaymode, "catch_fault_r", 13) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirCatchFault;
      else parsedMsg->playmode = PM_OurCatchFault;
    }
    else if ( strncmp( posPlaymode, "indirect_free_kick_l", 20) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurIndirectFreeKick;
      else parsedMsg->playmode = PM_TheirIndirectFreeKick;
    }
    else if ( strncmp( posPlaymode, "indirect_free_kick_r", 20) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirIndirectFreeKick;
      else parsedMsg->playmode = PM_OurIndirectFreeKick;
    }
    else if ( strncmp( posPlaymode, "penalty_setup_l", 15) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurPenaltySetup;
      else parsedMsg->playmode = PM_TheirIndirectFreeKick;
    }
    else if ( strncmp( posPlaymode, "penalty_setup_r", 15) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirIndirectFreeKick;
      else parsedMsg->playmode = PM_OurPenaltySetup;
    }
    else if ( strncmp( posPlaymode, "penalty_ready_l", 15) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurPenaltyReady;
      else parsedMsg->playmode = PM_TheirPenaltyReady;
    }
    else if ( strncmp( posPlaymode, "penalty_ready_r", 15) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirPenaltyReady;
      else parsedMsg->playmode = PM_OurPenaltyReady;
    }
    else if ( strncmp( posPlaymode, "penalty_taken_l", 15) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurPenaltyTaken;
      else parsedMsg->playmode = PM_TheirPenaltyTaken;
    }
    else if ( strncmp( posPlaymode, "penalty_taken_r", 15) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirPenaltyTaken;
      else parsedMsg->playmode = PM_OurPenaltyTaken;
    }
    else if ( strncmp( posPlaymode, "penalty_miss_l", 14) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurPenaltyMiss;
      else parsedMsg->playmode = PM_TheirPenaltyMiss;
    }
    else if ( strncmp( posPlaymode, "penalty_miss_r", 14) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirPenaltyMiss;
      else parsedMsg->playmode = PM_OurPenaltyMiss;
    }
    else if ( strncmp( posPlaymode, "penalty_score_l", 15) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_OurPenaltyScore;
      else parsedMsg->playmode = PM_TheirPenaltyScore;
    }
    else if ( strncmp( posPlaymode, "penalty_score_r", 15) == 0 ) {
      if ( ourSide == FS_Left ) parsedMsg->playmode = PM_TheirPenaltyScore;
      else parsedMsg->playmode = PM_OurPenaltyScore;
    }
    else parsedMsg->playmode = PM_UNKNOWN;
    
    return parsedMsg;
  }

  // hear player msg -> teammate or opponent said something brilliant
  HearPlayerMsg *parsedMsg = new HearPlayerMsg();
  sscanf( msg, "(hear %d", &(parsedMsg->serverTime) );
  char *pos = strchr( msg, 'p' );
  if ( pos != NULL ) {
    //if ( strncmp( pos + 3, ConfigFile::teamname.c_str(),
    if ( strncmp( pos + 3, teamname,
		  //strlen(ConfigFile::teamname.c_str()) ) == 0 ) {
      strlen(teamname) ) == 0 ) {
      parsedMsg->ourTeam = true;
    } else {
      parsedMsg->ourTeam = false;
    }
    // go to the closing quotation mark
    pos = strchr( pos + 3, '"' );
    sscanf( pos+1, " %d", &(parsedMsg->uNum) );
    // go to the opening quotation mark of the message
    pos = strchr( pos + 3, '"' );
    // get the position of the closing quotation mark
    char *posEnd = strchr( pos + 1, '"' );
    // copy the bytes between the quotation marks
    // would be nicer with strndup, but didn't compile :-(
    int bytesToCopy = posEnd - (pos+1);
    parsedMsg->message = (char *)malloc( bytesToCopy + 1 );    
    strncpy( parsedMsg->message, pos + 1, bytesToCopy );
    parsedMsg->message[ bytesToCopy ] = '\0';
  }
  return parsedMsg;
}



SeeGlobalMsg *Parser::parseSeeGlobalMsg( char *msg ) const {
  SeeGlobalMsg *parsedMsg = new SeeGlobalMsg();
  parseVisualData( msg, (VisualMsg *)parsedMsg );
  return parsedMsg;
}

OkLookMsg *Parser::parseOkLookMsg( char *msg ) const {
  OkLookMsg *parsedMsg = new OkLookMsg();
  parseVisualData( msg, (VisualMsg *)parsedMsg );
  return parsedMsg;
}


void Parser::swapSides(ParsedPlayer *p){
  if ( ourSide != FS_Right ) return;
  p->posX *= -1;
  p->posY *= -1;
  p->velX *= -1;
  p->velY *= -1;
  p->bodyAngle += 180;
  if (p->bodyAngle > 180){
    p->bodyAngle -= 360;
  }
  // neck-angle does not need to be changed since it is relative to
  // the bodyAngle
}


/* Parses a see_global or ok look message. Example:
   (see_global 75 ((g l) -52.5 0) ((g r) 52.5 0) 
   ((b) 5.11863 -23.9741 0 0) 
   ((p "vw2003" 1 goalie) -46.1879 -2.44873 2.84348e-11 -5.8194e-10 -113 89) 
   ((p "vw2003" 2) 3.02398 -5.52357 -5.4267e-09 -1.8675e-09 -85 2) 
   ...
   ((p "vw2003" 11) 3.7863 19.0476 2.66302e-08 -3.51868e-09 -90 2)) 
   ((p "enemy" 1 goalie) 2.24364 6.34819 -1.74014e-11 -2.3208e-10 -78 -5) 
   ...
   ((p "enemy" 11) -4.0312 -24.0305 0 0 -40 -40) 
*/
void Parser::parseVisualData( char *msg, VisualMsg *parsedMsg ) const {

  // calc our teamname length only one time
  //int nameLength = strlen( ConfigFile::teamname.c_str() );
  int nameLength = strlen( teamname );
  char *pos = msg;
  
  parsedMsg->serverTime = atoi( pos );

  // go to the next opening bracket
  pos = strchr( pos, '(' );
  // while not the end of the message is reached
  while ( pos != NULL ) {
    // switch over the char after "(("
    switch ( pos[ 2 ] ) {
    case 'p': 
      {
	ParsedPlayer *newPlayer = new ParsedPlayer();
	pos += 5;
	// player of our team
	//if ( strncmp( pos, ConfigFile::teamname.c_str(), nameLength ) == 0 ) {
	if ( strncmp( pos, teamname, nameLength ) == 0 ) {
	  parsedMsg->ourTeam->push_back( newPlayer );
	} 
	// player of their team 
	else {
	  parsedMsg->theirTeam->push_back( newPlayer );
	}

	// check if it is the goalie
	// look for the closing '"' of the teamname
	char *closingQuote = strchr( pos, '"' );
	if ( closingQuote != NULL ) {
	  char *nextG = strchr( closingQuote, 'g' );
	  if ( nextG != NULL &&
	       nextG - closingQuote <= 5 ) {
	    newPlayer->goalie = true;
	  }
	}

	if ( newPlayer->isGoalie() ) {
	  // get all the values at once (optional POINTING_DIRECTION is not used)
	  sscanf( closingQuote+1, // look after the teamname
		  "%d goalie) %f %f %f %f %f %f",
		  &(newPlayer->uNum), &(newPlayer->posX), &(newPlayer->posY),
		  &(newPlayer->velX), &(newPlayer->velY),
		  &(newPlayer->bodyAngle), &(newPlayer->neckAngle) );
	} else {
	  // get all the values at once (optional POINTING_DIRECTION is not used)
	  sscanf( closingQuote+1, // look after the teamname
		  "%d) %f %f %f %f %f %f",
		  &(newPlayer->uNum), &(newPlayer->posX), &(newPlayer->posY),
		  &(newPlayer->velX), &(newPlayer->velY),
		  &(newPlayer->bodyAngle), &(newPlayer->neckAngle) );
	  
	}
  if ( ourSide == FS_Right ){
	  newPlayer->posX *= -1;
	  newPlayer->posY *= -1;
	  newPlayer->velX *= -1;
	  newPlayer->velY *= -1;
    newPlayer->bodyAngle *= -1;
    newPlayer->neckAngle *= -1;
	}
	// go to the next opening bracket
	pos = strchr( closingQuote, '(' );
	break;
      }
    case 'b': // get the info of the ball
      {
	sscanf( pos+4 , "%f %f %f %f", 
		&(parsedMsg->ball->posX),
		&(parsedMsg->ball->posY),
		&(parsedMsg->ball->velX),
		&(parsedMsg->ball->velY) );
	// find the next opening bracket
	pos = strchr( pos+4, '(' );

	if ( ourSide == FS_Right ){
	  parsedMsg->ball->posX *= -1;
	  parsedMsg->ball->posY *= -1;
	  parsedMsg->ball->velX *= -1;
	  parsedMsg->ball->velY *= -1;
	}

	break;
      }
    case 'g':
      {
	// ignore goal part of the message
	pos = strchr( pos+4, '(' );
	break;
      }
    default:
      {
	DEBUG2( "ERROR: Parser::parseVisualData: Did not understand symbol: ", pos[ 2 ] );
	pos = strchr( pos+4, '(' );
      }
    }
  }
}

TeamGraphicOkMsg *Parser::parseTeamGraphicOkMsg( char *msg ) const{
  int x,y;  
  sscanf(msg, "(ok team_graphic %i %i)", &x, &y);
  //DEBUG6("parseTeamGraphicOkMsg: ", msg, " ", x, " ", y);
  TeamGraphicOkMsg *tmp = new TeamGraphicOkMsg();
  tmp->x = x;
  tmp->y = y;
  return tmp;
}

/* typical format:
   (ok team_names (team l vw2003) (team r trilearn))
*/
OkTeamnamesMsg *Parser::parseOkTeamnamesMsg( char *msg ) const {
  OkTeamnamesMsg *parsedMsg = new OkTeamnamesMsg();

  // search for a left teamname
  char *posFirst = strstr( msg, "(team l " );
  if ( posFirst != NULL ) {
    // go to the first char of the teamname
    posFirst = posFirst + 8;
    // find the last char of the teamname
    char *posLast = strchr( posFirst, ')' );
    if ( posLast != NULL ) {
      // copy the teamname to the parsedMsg
      int bytesToCopy = posLast - posFirst;
      parsedMsg->teamnameLeft = (char *)malloc( bytesToCopy + 1 );    
      strncpy( parsedMsg->teamnameLeft, posFirst, bytesToCopy );
      parsedMsg->teamnameLeft[ bytesToCopy ] = '\0';
    }
  }
  
  // search for a right teamname
  posFirst = strstr( msg, "(team r " );
  if ( posFirst != NULL ) {
    // go to the first char of the teamname
    posFirst = posFirst + 8;
    // find the last char of the teamname
    char *posLast = strchr( posFirst, ')' );
    if ( posLast != NULL ) {
      // copy the teamname to the parsedMsg
      int bytesToCopy = posLast - posFirst;
      parsedMsg->teamnameRight = (char *)malloc( bytesToCopy + 1 );    
      strncpy( parsedMsg->teamnameRight, posFirst, bytesToCopy );
      parsedMsg->teamnameRight[ bytesToCopy ] = '\0';
    }
  }

  return parsedMsg;
}



/* typical format:
   (clang (ver (p "vw2003" 1) 7 7))
*/
CLangVersPlayerMsg *Parser::parseCLangVersPlayerMsg( char *msg ) const {
  CLangVersPlayerMsg *parsedMsg = new CLangVersPlayerMsg();
  // our player?
  //if ( strncmp( msg, ConfigFile::teamname.c_str(),
  if ( strncmp( msg, teamname,
	// strlen(ConfigFile::teamname.c_str()) ) == 0 ) {
    strlen(teamname) ) == 0 ) {
    parsedMsg->ourTeam = true;
  } else {
    parsedMsg->ourTeam = false;
  }
  char *pos = strchr( msg, '"' );
  if ( pos != NULL ) {
    sscanf( pos + 2, "%d) %d %d", &(parsedMsg->uNum), 
	    &(parsedMsg->minCLangVers), &(parsedMsg->maxCLangVers) );
  }
  return parsedMsg;
}

/* typical format:
   (ok change_player_type 2 3)
*/
OkChangePlayerTypeMsg *Parser::parseOkChangePlayerTypeMsg( char *msg ) const {
  OkChangePlayerTypeMsg *parsedMsg = new OkChangePlayerTypeMsg();
  sscanf( msg, "%d %d", &(parsedMsg->uNum), &(parsedMsg->playerType) );
  return parsedMsg;
}

ChangePlayerTypeMsg *Parser::parseChangePlayerTypeMsg ( char *msg ) const {
  ChangePlayerTypeMsg *parsedMsg = new ChangePlayerTypeMsg();
  sscanf( msg, "%d", &(parsedMsg->uNum));
  return parsedMsg; 
}

/* typical format:
   (player_type (id 1) (player_speed_max 1.2) (stamina_inc_max 28) 
   (player_decay 0.4972) (inertia_moment 7.43) (dash_power_rate 0.0077) 
   (player_size 0.3) (kickable_margin 0.8878) (kick_rand 0.0939) 
   (extra_stamina 81.9) (effort_max 0.8362) (effort_min 0.4362))
*/
PlayerTypeMsg *Parser::parsePlayerTypeMsg( char *msg ) const {
  PlayerTypeMsg *parsedMsg = new PlayerTypeMsg();
  char *pos = NULL;

  pos = strstr( msg, "(id " );
  if ( pos != NULL ) parsedMsg->id = atoi( pos + 4 );
  
  pos = strstr( msg, "(player_speed_max " );
  if ( pos != NULL ) parsedMsg->player_speed_max = atof( pos + 18 );
    
  pos = strstr( msg, "(stamina_inc_max " );
  if ( pos != NULL ) parsedMsg->stamina_inc_max = atof( pos + 17 );

  pos = strstr( msg, "(player_decay " );
  if ( pos != NULL ) parsedMsg->player_decay = atof( pos + 14 );

  pos = strstr( msg, "(inertia_moment " );
  if ( pos != NULL ) parsedMsg->inertia_moment = atof( pos + 16 );

  pos = strstr( msg, "(dash_power_rate " );
  if ( pos != NULL ) parsedMsg->dash_power_rate = atof( pos + 17 );

  pos = strstr( msg, "(player_size " );
  if ( pos != NULL ) parsedMsg->player_size = atof( pos + 13 );

  pos = strstr( msg, "(kickable_margin " );
  if ( pos != NULL ) parsedMsg->kickable_margin = atof( pos + 17 );

  pos = strstr( msg, "(kick_rand " );
  if ( pos != NULL ) parsedMsg->kick_rand = atof( pos + 11 );

  pos = strstr( msg, "(extra_stamina " );
  if ( pos != NULL ) parsedMsg->extra_stamina = atof( pos + 15 );

  pos = strstr( msg, "(effort_max " );
  if ( pos != NULL ) parsedMsg->effort_max = atof( pos + 12 );

  pos = strstr( msg, "(effort_min " );
  if ( pos != NULL ) parsedMsg->effort_min = atof( pos + 12 );

  return parsedMsg;
}

/* typical message:
   (player_param (player_types 7)(pt_max 3)(random_seed -1)(subs_max 3)
   (dash_power_rate_delta_max 0)(dash_power_rate_delta_min 0)
   (effort_max_delta_factor -0.002)(effort_min_delta_factor -0.002)
   (extra_stamina_delta_max 100)(extra_stamina_delta_min 0)
   (inertia_moment_delta_factor 25)(kick_rand_delta_factor 0.5)
   (kickable_margin_delta_max 0.2)(kickable_margin_delta_min 0)
   (new_dash_power_rate_delta_max 0.002)(new_dash_power_rate_delta_min 0)
   (new_stamina_inc_max_delta_factor -10000)
   (player_decay_delta_max 0.2)(player_decay_delta_min 0)
   (player_size_delta_factor -100)(player_speed_max_delta_max 0)
   (player_speed_max_delta_min 0)(stamina_inc_max_delta_factor 0))
*/
PlayerParamMsg *Parser::parsePlayerParamMsg( char *msg ) const {
  PlayerParamMsg *parsedMsg = new PlayerParamMsg();
  char *pos = NULL;

  pos = strstr( msg, "(player_types " );
  if ( pos != NULL ) parsedMsg->player_types = atoi( pos + 14 );
  pos = strstr( msg, "(pt_max " );
  if ( pos != NULL ) parsedMsg->pt_max = atoi( pos + 8 );
  pos = strstr( msg, "(random_seed " );
  if ( pos != NULL ) parsedMsg->random_seed = atoi( pos + 13 );
  pos = strstr( msg, "(subs_max " );
  if ( pos != NULL ) parsedMsg->subs_max = atoi( pos + 10 );

  pos = strstr( msg, "(dash_power_rate_delta_max " );
  if ( pos != NULL ) parsedMsg->dash_power_rate_delta_max = atof( pos + 27 );
  pos = strstr( msg, "(dash_power_rate_delta_min " );
  if ( pos != NULL ) parsedMsg->dash_power_rate_delta_min = atof( pos + 27 );
  pos = strstr( msg, "(effort_max_delta_factor " );
  if ( pos != NULL ) parsedMsg->effort_max_delta_factor = atof( pos + 25 );
  pos = strstr( msg, "(effort_min_delta_factor " );
  if ( pos != NULL ) parsedMsg->effort_min_delta_factor = atof( pos + 25 );
  pos = strstr( msg, "(extra_stamina_delta_max " );
  if ( pos != NULL ) parsedMsg->extra_stamina_delta_max = atof( pos + 25 );
  pos = strstr( msg, "(extra_stamina_delta_min " );
  if ( pos != NULL ) parsedMsg->extra_stamina_delta_min = atof( pos + 25 );
  pos = strstr( msg, "(inertia_moment_delta_factor " );
  if ( pos != NULL ) parsedMsg->inertia_moment_delta_factor = atof( pos + 29 );
  pos = strstr( msg, "(kick_rand_delta_factor " );
  if ( pos != NULL ) parsedMsg->kick_rand_delta_factor = atof( pos + 24 );
  pos = strstr( msg, "(kickable_margin_delta_max " );
  if ( pos != NULL ) parsedMsg->kickable_margin_delta_max = atof( pos + 27 );
  pos = strstr( msg, "(kickable_margin_delta_min " );
  if ( pos != NULL ) parsedMsg->kickable_margin_delta_min = atof( pos + 27 );
  pos = strstr( msg, "(new_dash_power_rate_delta_max " );
  if ( pos != NULL ) parsedMsg->new_dash_power_rate_delta_max = atof( pos + 31 );
  pos = strstr( msg, "(new_dash_power_rate_delta_min " );
  if ( pos != NULL ) parsedMsg->new_dash_power_rate_delta_min = atof( pos + 31 );
  pos = strstr( msg, "(new_stamina_inc_max_delta_factor " );
  if ( pos != NULL ) parsedMsg->new_stamina_inc_max_delta_factor = atof( pos + 34 );
  pos = strstr( msg, "(player_decay_delta_max " );
  if ( pos != NULL ) parsedMsg->player_decay_delta_max = atof( pos + 24 );
  pos = strstr( msg, "(player_decay_delta_min " );
  if ( pos != NULL ) parsedMsg->player_decay_delta_min = atof( pos + 24 );
  pos = strstr( msg, "(player_size_delta_factor " );
  if ( pos != NULL ) parsedMsg->player_size_delta_factor = atof( pos + 26 );
  pos = strstr( msg, "(player_speed_max_delta_max " );
  if ( pos != NULL ) parsedMsg->player_speed_max_delta_max = atof( pos + 28 );
  pos = strstr( msg, "(player_speed_max_delta_min " );
  if ( pos != NULL ) parsedMsg->player_speed_max_delta_min = atof( pos + 28 );
  pos = strstr( msg, "(stamina_inc_max_delta_factor " );
  if ( pos != NULL ) parsedMsg->stamina_inc_max_delta_factor = atof( pos + 30 );

  return parsedMsg;
}


/* typical format of this beast:
   (server_param (audio_cut_dist 50)(back_passes 1)(ball_accel_max 2.7)
   (ball_decay 0.94)(ball_rand 0.05)(ball_size 0.085)(ball_speed_max 2.7)
   (ball_weight 0.2)(catch_ban_cycle 5)(catch_probability 1)(catchable_area_l 2)
   (catchable_area_w 1)(ckick_margin 1)(clang_advice_win 1)(clang_define_win 1)
   (clang_del_win 1)(clang_info_win 1)(clang_mess_delay 50)(clang_mess_per_cycle 1)
   (clang_meta_win 1)(clang_rule_win 1)(clang_win_size 300)(coach 0)(coach_port 6001)
   (coach_w_referee 0)(control_radius 2)(dash_power_rate 0.006)(drop_ball_time 200)
   (effort_dec 0.005)(effort_dec_thr 0.3)(effort_inc 0.01)(effort_inc_thr 0.6)
   (effort_init 1)(effort_min 0.6)(forbid_kick_off_offside 1)(free_kick_faults 1)
   (freeform_send_period 20)(freeform_wait_period 600)(fullstate_l 0)(fullstate_r 0)
   (game_log_compression 0)(game_log_dated 1)(game_log_dir "./")(game_log_fixed 0)
   (game_log_fixed_name "rcssserver")(game_log_version 3)(game_logging 1)
   (goal_width 14.02)(goalie_max_moves 2)(half_time 3000)(hear_decay 1)(hear_inc 1)
   (hear_max 1)(inertia_moment 5)(kick_power_rate 0.027)(kick_rand 0)
   (kick_rand_factor_l 1)(kick_rand_factor_r 1)(kickable_margin 0.7)
   (landmark_file "~/.rcssserver-landmark.xml")(log_date_format "%Y%m%d%H%M-")
   (log_times 0)(max_goal_kicks 3)(maxmoment 180)(maxneckang 90)(maxneckmoment 180)
   (maxpower 100)(minmoment -180)(minneckang -90)(minneckmoment -180)(minpower -100)
   (offside_active_area_size 2.5)(offside_kick_margin 9.15)(olcoach_port 6002)
   (old_coach_hear 0)(player_accel_max 1)(player_decay 0.4)(player_rand 0.1)
   (player_size 0.3)(player_speed_max 1.2)(player_weight 60)(point_to_ban 5)
   (point_to_duration 20)(port 6000)(prand_factor_l 1)(prand_factor_r 1)(profile 0)
   (proper_goal_kicks 0)(quantize_step 0.1)(quantize_step_l 0.01)(record_messages 0)
   (recover_dec 0.002)(recover_dec_thr 0.3)(recover_min 0.5)(recv_step 10)
   (say_coach_cnt_max 128)(say_coach_msg_size 128)(say_msg_size 10)(send_comms 0)
   (send_step 150)(send_vi_step 100)(sense_body_step 100)(simulator_step 100)
   (slow_down_factor 1)(slowness_on_top_for_left_team 1)
   (slowness_on_top_for_right_team 1)(stamina_inc_max 45)(stamina_max 4000)
   (start_goal_l 0)(start_goal_r 0)(stopped_ball_vel 0.01)(synch_micro_sleep 1)
   (synch_mode 0)(synch_offset 60)(tackle_back_dist 0.5)(tackle_cycles 10)
   (tackle_dist 2)(tackle_exponent 6)(tackle_power_rate 0.027)(tackle_width 1)
   (team_actuator_noise 0)(text_log_compression 0)(text_log_dated 1)
   (text_log_dir "./")(text_log_fixed 0)(text_log_fixed_name "rcssserver")
   (text_logging 1)(use_offside 1)(verbose 0)(visible_angle 90)(visible_distance 3)
   (wind_ang 0)(wind_dir 0)(wind_force 0)(wind_none 0)(wind_rand 0)(wind_random 0))
*/
ServerParamMsg *Parser::parseServerParamMsg( char *msg ) const {
  ServerParamMsg *parsedMsg = new ServerParamMsg();
  char *pos = NULL;

  pos = strstr( msg, "(audio_cut_dist " );
  if ( pos != NULL ) parsedMsg->audio_cut_dist = atof( pos + 16 );
  pos = strstr( msg, "(back_passes " );
  if ( pos != NULL ) parsedMsg->back_passes = atoi( pos + 13 );
  pos = strstr( msg, "(ball_accel_max " );
  if ( pos != NULL ) parsedMsg->ball_accel_max = atof( pos + 16 );
  pos = strstr( msg, "(ball_decay " );
  if ( pos != NULL ) parsedMsg->ball_decay = atof( pos + 12 );
  pos = strstr( msg, "(ball_rand " );
  if ( pos != NULL ) parsedMsg->ball_rand = atof( pos + 11 );
  pos = strstr( msg, "(ball_size " );
  if ( pos != NULL ) parsedMsg->ball_size = atof( pos + 11 );
  pos = strstr( msg, "(ball_speed_max " );
  if ( pos != NULL ) parsedMsg->ball_speed_max = atof( pos + 16 );
  pos = strstr( msg, "(ball_weight " );
  if ( pos != NULL ) parsedMsg->ball_weight = atof( pos + 13 );
  pos = strstr( msg, "(catch_probability " );
  if ( pos != NULL ) parsedMsg->catch_probability = atof( pos + 19 );
  pos = strstr( msg, "(catchable_area_l " );
  if ( pos != NULL ) parsedMsg->catchable_area_l = atof( pos + 18 );
  pos = strstr( msg, "(catchable_area_w " );
  if ( pos != NULL ) parsedMsg->catchable_area_w = atof( pos + 18 );
  pos = strstr( msg, "(ckick_margin " );
  if ( pos != NULL ) parsedMsg->ckick_margin = atof( pos + 14 );
  pos = strstr( msg, "(clang_advice_win " );
  if ( pos != NULL ) parsedMsg->clang_advice_win = atoi( pos + 18 );
  pos = strstr( msg, "(clang_define_win " );
  if ( pos != NULL ) parsedMsg->clang_define_win = atoi( pos + 18 );
  pos = strstr( msg, "(clang_del_win " );
  if ( pos != NULL ) parsedMsg->clang_del_win = atoi( pos + 15 );
  pos = strstr( msg, "(clang_info_win " );
  if ( pos != NULL ) parsedMsg->clang_info_win = atoi( pos + 16 );
  pos = strstr( msg, "(clang_mess_delay " );
  if ( pos != NULL ) parsedMsg->clang_mess_delay = atoi( pos + 18 );
  pos = strstr( msg, "(clang_mess_per_cycle " );
  if ( pos != NULL ) parsedMsg->clang_mess_per_cycle = atoi( pos + 22 );
  pos = strstr( msg, "(clang_meta_win " );
  if ( pos != NULL ) parsedMsg->clang_meta_win = atoi( pos + 16 );
  pos = strstr( msg, "(clang_rule_win " );
  if ( pos != NULL ) parsedMsg->clang_rule_win = atoi( pos + 16 );
  pos = strstr( msg, "(clang_win_size " );
  if ( pos != NULL ) parsedMsg->clang_win_size = atoi( pos + 16 );
  pos = strstr( msg, "(dash_power_rate " );
  if ( pos != NULL ) parsedMsg->dash_power_rate = atof( pos + 17 );
  pos = strstr( msg, "(drop_ball_time " );
  if ( pos != NULL ) parsedMsg->drop_ball_time = atoi( pos + 16 );
  pos = strstr( msg, "(effort_dec " );
  if ( pos != NULL ) parsedMsg->effort_dec = atof( pos + 12 );
  pos = strstr( msg, "(effort_dec_thr " );
  if ( pos != NULL ) parsedMsg->effort_dec_thr = atof( pos + 16 );
  pos = strstr( msg, "(effort_inc " );
  if ( pos != NULL ) parsedMsg->effort_inc = atof( pos + 12 );
  pos = strstr( msg, "(effort_inc_thr " );
  if ( pos != NULL ) parsedMsg->effort_inc_thr = atof( pos + 16 );
  pos = strstr( msg, "(effort_init " );
  if ( pos != NULL ) parsedMsg->effort_init = atof( pos + 13 );
  pos = strstr( msg, "(effort_min " );
  if ( pos != NULL ) parsedMsg->effort_min = atof( pos + 12 );
  pos = strstr( msg, "(forbid_kick_off_offside " );
  if ( pos != NULL ) parsedMsg->forbid_kick_off_offside = atoi(pos + 25 );
  pos = strstr( msg, "(free_kick_faults " );
  if ( pos != NULL ) parsedMsg->free_kick_faults = atoi( pos + 18 );
  pos = strstr( msg, "(freeform_send_period " );
  if ( pos != NULL ) parsedMsg->freeform_send_period = atoi( pos + 22 );
  pos = strstr( msg, "(freeform_wait_period " );
  if ( pos != NULL ) parsedMsg->freeform_wait_period = atoi( pos + 22 );
  pos = strstr( msg, "(fullstate_l " );
  if ( pos != NULL ) parsedMsg->fullstate_l = atoi( pos + 13 );
  pos = strstr( msg, "(fullstate_r " );
  if ( pos != NULL ) parsedMsg->fullstate_r = atoi( pos + 13 );
  pos = strstr( msg, "(goal_width " );
  if ( pos != NULL ) parsedMsg->goal_width = atof( pos + 12 );
  pos = strstr( msg, "(goalie_max_moves " );
  if ( pos != NULL ) parsedMsg->goalie_max_moves = atoi( pos + 18 );
  pos = strstr( msg, "(half_time " );
  if ( pos != NULL ) parsedMsg->half_time = atoi( pos + 11 );
  pos = strstr( msg, "(hear_decay " );
  if ( pos != NULL ) parsedMsg->hear_decay = atoi( pos + 13 );
  pos = strstr( msg, "(hear_inc " );
  if ( pos != NULL ) parsedMsg->hear_inc = atoi( pos + 10 );
  pos = strstr( msg, "(hear_max " );
  if ( pos != NULL ) parsedMsg->hear_max = atoi( pos + 10 );
  pos = strstr( msg, "(inertia_moment " );
  if ( pos != NULL ) parsedMsg->inertia_moment = atof( pos + 16 );
  pos = strstr( msg, "(kick_power_rate " );
  if ( pos != NULL ) parsedMsg->kick_power_rate = atof( pos + 17 );
  pos = strstr( msg, "(kick_rand " );
  if ( pos != NULL ) parsedMsg->kick_rand = atof( pos + 11 );
  pos = strstr( msg, "(kick_rand_factor_l " );
  if ( pos != NULL ) parsedMsg->kick_rand_factor_l = atof( pos + 20 );
  pos = strstr( msg, "(kick_rand_factor_r " );
  if ( pos != NULL ) parsedMsg->kick_rand_factor_r = atof( pos + 20 );
  pos = strstr( msg, "(kickable_margin " );
  if ( pos != NULL ) parsedMsg->kickable_margin = atof( pos + 17 );
  pos = strstr( msg, "(max_goal_kicks " );
  if ( pos != NULL ) parsedMsg->max_goal_kicks = atoi( pos + 16 );
  pos = strstr( msg, "(maxmoment " );
  if ( pos != NULL ) parsedMsg->maxmoment = atof( pos + 11 );
  pos = strstr( msg, "(maxneckang " );
  if ( pos != NULL ) parsedMsg->maxneckang = atof( pos + 12 );
  pos = strstr( msg, "(maxneckmoment " );
  if ( pos != NULL ) parsedMsg->maxneckmoment = atof( pos + 15 );
  pos = strstr( msg, "(maxpower " );
  if ( pos != NULL ) parsedMsg->maxpower = atof( pos + 10 );
  pos = strstr( msg, "(minmoment " );
  if ( pos != NULL ) parsedMsg->minmoment = atof( pos + 11 );
  pos = strstr( msg, "(minneckang " );
  if ( pos != NULL ) parsedMsg->minneckang = atof( pos + 12 );
  pos = strstr( msg, "(minneckmoment " );
  if ( pos != NULL ) parsedMsg->minneckmoment = atof( pos + 15 );
  pos = strstr( msg, "(minpower " );
  if ( pos != NULL ) parsedMsg->minpower = atof( pos + 10 );
  pos = strstr( msg, "(offside_active_area_size " );
  if ( pos != NULL ) parsedMsg->offside_active_area_size = atof( pos + 26 );
  pos = strstr( msg, "(offside_kick_margin " );
  if ( pos != NULL ) parsedMsg->offside_kick_margin = atof( pos + 21 );
  pos = strstr( msg, "(player_accel_max " );
  if ( pos != NULL ) parsedMsg->player_accel_max = atof( pos + 18 );
  pos = strstr( msg, "(player_decay " );
  if ( pos != NULL ) parsedMsg->player_decay = atof( pos + 14 );
  pos = strstr( msg, "(player_rand " );
  if ( pos != NULL ) parsedMsg->player_rand = atof( pos + 13 );
  pos = strstr( msg, "(player_size " );
  if ( pos != NULL ) parsedMsg->player_size = atof( pos + 13 );
  pos = strstr( msg, "(player_speed_max " );
  if ( pos != NULL ) parsedMsg->player_speed_max = atof( pos + 18 );
  pos = strstr( msg, "(player_weight " );
  if ( pos != NULL ) parsedMsg->player_weight = atof( pos + 15 );
  pos = strstr( msg, "(point_to_ban " );
  if ( pos != NULL ) parsedMsg->point_to_ban = atoi( pos + 14 );
  pos = strstr( msg, "(point_to_duration " );
  if ( pos != NULL ) parsedMsg->point_to_duration = atoi( pos + 19 );
  pos = strstr( msg, "(prand_factor_l " );
  if ( pos != NULL ) parsedMsg->prand_factor_l = atof( pos + 16 );
  pos = strstr( msg, "(prand_factor_r " );
  if ( pos != NULL ) parsedMsg->prand_factor_r = atof( pos + 16 );
  pos = strstr( msg, "(proper_goal_kicks " );
  if ( pos != NULL ) parsedMsg->proper_goal_kicks = atoi( pos + 19 );
  pos = strstr( msg, "(quantize_step " );
  if ( pos != NULL ) parsedMsg->quantize_step = atof( pos + 15 );
  pos = strstr( msg, "(quantize_step_l " );
  if ( pos != NULL ) parsedMsg->quantize_step_l = atof( pos + 17 );
  pos = strstr( msg, "(recover_dec " );
  if ( pos != NULL ) parsedMsg->recover_dec = atof( pos + 13 );
  pos = strstr( msg, "(recover_dec_thr " );
  if ( pos != NULL ) parsedMsg->recover_dec_thr = atof( pos + 17 );
  pos = strstr( msg, "(recover_min " );
  if ( pos != NULL ) parsedMsg->recover_min = atof( pos + 13 );
  pos = strstr( msg, "(recv_step " );
  if ( pos != NULL ) parsedMsg->recv_step = atof( pos + 11 );
  pos = strstr( msg, "(say_coach_cnt_max " );
  if ( pos != NULL ) parsedMsg->say_coach_cnt_max = atoi( pos + 19 );
  pos = strstr( msg, "(say_coach_msg_size " );
  if ( pos != NULL ) parsedMsg->say_coach_msg_size = atoi( pos + 20 );
  pos = strstr( msg, "(say_msg_size " );
  if ( pos != NULL ) parsedMsg->say_msg_size = atoi( pos + 14 );
  pos = strstr( msg, "(send_comms " );
  if ( pos != NULL ) parsedMsg->send_comms = atoi( pos + 12 );
  pos = strstr( msg, "(send_step " );
  if ( pos != NULL ) parsedMsg->send_step = atoi( pos + 11 );
  pos = strstr( msg, "(send_vi_step " );
  if ( pos != NULL ) parsedMsg->send_vi_step = atoi( pos + 14 );
  pos = strstr( msg, "(sense_body_step " );
  if ( pos != NULL ) parsedMsg->sense_body_step = atoi( pos + 17 );
  pos = strstr( msg, "(simulator_step " );
  if ( pos != NULL ) parsedMsg->simulator_step = atoi( pos + 16 );
  pos = strstr( msg, "(slow_down_factor " );
  if ( pos != NULL ) parsedMsg->slow_down_factor = atof( pos + 18 );
  pos = strstr( msg, "(slowness_on_top_for_left_team " );
  if ( pos != NULL ) parsedMsg->slowness_on_top_for_left_team = atoi( pos + 31 );
  pos = strstr( msg, "(slowness_on_top_for_right_team " );
  if ( pos != NULL ) parsedMsg->slowness_on_top_for_right_team = atoi( pos + 32 );
  pos = strstr( msg, "(stamina_inc_max " );
  if ( pos != NULL ) parsedMsg->stamina_inc_max = atof( pos + 17 );
  pos = strstr( msg, "(stamina_max " );
  if ( pos != NULL ) parsedMsg->stamina_max = atof( pos + 13 );
  pos = strstr( msg, "(start_goal_l " );
  if ( pos != NULL ) parsedMsg->start_goal_l = atoi( pos + 14 );
  pos = strstr( msg, "(start_goal_r " );
  if ( pos != NULL ) parsedMsg->start_goal_r = atoi( pos + 14 );
  pos = strstr( msg, "(stopped_ball_vel " );
  if ( pos != NULL ) parsedMsg->stopped_ball_vel = atof( pos + 18 );
  pos = strstr( msg, "(synch_micro_sleep " );
  if ( pos != NULL ) parsedMsg->synch_micro_sleep = atoi( pos + 19 );
  pos = strstr( msg, "(synch_mode " );
  if ( pos != NULL ) parsedMsg->synch_mode = atoi( pos + 12 );
  pos = strstr( msg, "(synch_offset " );
  if ( pos != NULL ) parsedMsg->synch_offset = atoi( pos + 14 );
  pos = strstr( msg, "(tackle_back_dist " );
  if ( pos != NULL ) parsedMsg->tackle_back_dist = atof( pos + 18 );
  pos = strstr( msg, "(tackle_cycles " );
  if ( pos != NULL ) parsedMsg->tackle_cycles = atoi( pos + 15 );
  pos = strstr( msg, "(tackle_dist " );
  if ( pos != NULL ) parsedMsg->tackle_dist = atof( pos + 13 );
  pos = strstr( msg, "(tackle_exponent " );
  if ( pos != NULL ) parsedMsg->tackle_exponent = atof( pos + 17 );
  pos = strstr( msg, "(tackle_power_rate " );
  if ( pos != NULL ) parsedMsg->tackle_power_rate = atof( pos + 19 );
  pos = strstr( msg, "(tackle_width " );
  if ( pos != NULL ) parsedMsg->tackle_width = atof( pos + 14 );
  pos = strstr( msg, "(use_offside " );
  if ( pos != NULL ) parsedMsg->use_offside = atoi( pos + 13 );
  pos = strstr( msg, "(visible_angle " );
  if ( pos != NULL ) parsedMsg->visible_angle = atof( pos + 15 );
  pos = strstr( msg, "(visible_distance " );
  if ( pos != NULL ) parsedMsg->visible_distance = atof( pos + 18 );
  pos = strstr( msg, "(wind_ang " );
  if ( pos != NULL ) parsedMsg->wind_ang = atof( pos + 10 );
  pos = strstr( msg, "(wind_dir " );
  if ( pos != NULL ) parsedMsg->wind_dir = atof( pos + 10 );
  pos = strstr( msg, "(wind_force " );
  if ( pos != NULL ) parsedMsg->wind_force = atof( pos + 12 );
  pos = strstr( msg, "(wind_none " );
  if ( pos != NULL ) parsedMsg->wind_none = atoi( pos + 11 );
  pos = strstr( msg, "(wind_rand " );
  if ( pos != NULL ) parsedMsg->wind_rand = atof( pos + 11 );
  pos = strstr( msg, "(wind_random " );
  if ( pos != NULL ) parsedMsg->wind_random = atoi( pos + 13 );

  return parsedMsg;
}
