/******************************************************************************
 * Communicator.cc
 * Copyright (c) 2003 Virtual Werder
 *
 *
 * implements:
 *   class Communicator
 *
 * requires:
 *
 *
 * Created on: 07.05.2003 by pagra
 * Maintainer: pagra
 * Group:
 * Co-authors: josch
 *****************************************************************************/


#include "Communicator.hh"
#include "debug.hh"


Communicator::Communicator() {
}

Communicator::~Communicator() {
  ;
}



void Communicator::closeConnection() {
  close( m_socket.socketFd );
}


bool Communicator::sendMessage( const char *msg ) {
  int length = strlen( msg ) + 1;  // + 1 for '\0'-termination
  DEBUG2( "Sending: ", msg );
  if ( sendto( m_socket.socketFd, msg, length, 0,
	       (struct sockaddr *)&m_socket.serv_addr, 
	       sizeof(m_socket.serv_addr)) 
       != length ) {
    DEBUG2("Sending message failed:\n", msg);
    return false;
  }
  return true;
}


/*
 * param buf, buffer with 65535 bytes to be filled with the message
 *
 * return false: server is dead
 *        true: message arrived
 */
bool Communicator::waitForAndReadMessage( char *buf ) {
  // paranoia: terminate
  buf[0] = '\0';

  int bytesRead = -1;
  int counter = 0;
  while ( bytesRead == -1 && counter < 50 ) {
    // try to read message
    bytesRead = recv( m_socket.socketFd, buf, 65534, 0 ); // nonblocking

    // no message to read -> sleep a bit
    if ( bytesRead == -1 ) {
      usleep( 100000 );
      counter++;
    } else {
      // terminate
      buf[bytesRead] = '\0';
    }
  }
  
  // server is dead (>5s no message)
  return ( bytesRead != -1 );
}




void Communicator::openConnection( const char *host, int port, 
				   const char *teamname, float version,
				   char *buf ) {

  m_socket.socketFd = -1;
  

  // host entity: infos about the host
  struct hostent *he;

  if ( (he = gethostbyname(host)) == NULL) {  // get the host info
    DEBUG1("Communicator::openConnection: can't get host by name");
    exit(1);
  }

  // open UDP socket
  if ( (m_socket.socketFd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    DEBUG1("Communicator::openConnection: can't get socket file descriptor");
    exit(1);
  }


  struct sockaddr_in my_addr;

  my_addr.sin_family = AF_INET;         // host byte order
  //  my_addr.sin_port = htons(port);       // short, network byte order
  my_addr.sin_port = htons( 0 );
  //  my_addr.sin_addr.s_addr = INADDR_ANY; // automatically fill with my IP
  my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  memset(&(my_addr.sin_zero), '\0', 8); // zero the rest of the struct


  // make socket non blocking on read operations
  int val; // container for the flags
  // read the flags, that are set for our socket
  val = fcntl(m_socket.socketFd, F_GETFL, 0);
  // add the flag O_NDELAY: the socket will not block on any operation
  val |= O_NDELAY;
  // set the flags for our socket
  fcntl(m_socket.socketFd, F_SETFL, val) ;

  // bind the socket to the desired port
  if ( bind( m_socket.socketFd, (struct sockaddr *)&my_addr,
	     sizeof(struct sockaddr)) 
       == -1 ) {
    DEBUG2("Communicator::openConnection: can't bind socket to port ", port);
    exit(1);
  }

  // UDP Socket is open for receiving


  // send init

  m_socket.serv_addr.sin_family = AF_INET;     // host byte order
  m_socket.serv_addr.sin_port = htons(port);   // short, network byte order
  m_socket.serv_addr.sin_addr = *((struct in_addr *)he->h_addr);
  memset(&(m_socket.serv_addr.sin_zero), '\0', 8); // zero the rest of the struct

  sprintf( buf, "(init %s (version %.1f))", teamname, version );
  int messagelength = strlen(buf) + 1;
  DEBUG2("Sending: ", buf);
  if ( sendto( m_socket.socketFd, buf, messagelength, 0,
	       (struct sockaddr *)&m_socket.serv_addr, 
	       sizeof(m_socket.serv_addr)) 
       != messagelength ) {
    DEBUG1("Communicator::openConnection: Sending init message: failed");
    exit(1);
  }


  // wait until recvfrom received a message;
  // then save the serverport

  buf[0] = '\0';  // reset buffer
  socklen_t servlen;
  struct sockaddr_in serv_addr;
  servlen = sizeof(serv_addr);
  int receivedBytes = -1; // nothing received yet
  int counter = 0; // first try
  // while nothing received and we tried for less than 200 times
  while ( receivedBytes == -1 && counter < 200 ) {
    // Try to receive message (nonblocking). The message is filled in buf and
    // the serverport is filled in serv_addr on success. Else receivedBytes will be -1.
    receivedBytes = recvfrom( m_socket.socketFd, 
			      (char *)buf, 65534, 0,
			      (struct sockaddr *)&serv_addr,
			      &servlen );
    // sleep a litte, if no message arrived yet
    if ( receivedBytes == -1 ) {
      usleep( 20000 );
      counter++;
    }
  }
  if ( receivedBytes < 0 ) {
    // we waited 200 * 0.02 = 4s -> no message arrived -> server dead?
    DEBUG1("wait for init ok: error while receiving. Probably no server.");
    exit(1);
  } else {
    // Paranoia: terminate the message
    buf[receivedBytes] = '\0';
    // save the server-port
    m_socket.serv_addr.sin_port = serv_addr.sin_port;
  }

  DEBUG2("received: \n", buf);

  DEBUG1("end openConnection()");
}















/////////////// Alter Mll - wird demnchst gelscht


  /*
  struct sockaddr_in my_addr;
  
  // UDP-Socket ffnen (incoming)
  if ((socketFdIn = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    cout << "Fehler";
    exit(1);
  }
  
  my_addr.sin_family = AF_INET;         // host byte order
  my_addr.sin_port = htons(port);       // short, network byte order
  my_addr.sin_addr.s_addr = INADDR_ANY; // automatically fill with my IP
  memset(&(my_addr.sin_zero), '\0', 8); // zero the rest of the struct
  
  // socket auf gewnschten Port binden
  if (bind(socketFdIn, (struct sockaddr *)&my_addr,
	   sizeof(struct sockaddr)) == -1) {
    cout << "Fehler bei bind" << endl;
    exit(1);
  }



  // UDP-Socket ffnen (outgoing)
  struct hostent *he;

  if ((he=gethostbyname(host)) == NULL) {  // get the host info
    cout << "gethostbyname" << endl;
    exit(1);
  }
  if ((socketFdOut = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    cout << "socket" << endl;
    exit(1);
  }
  
  their_addr.sin_family = AF_INET;     // host byte order
  their_addr.sin_port = htons(port); // short, network byte order
  their_addr.sin_addr = *((struct in_addr *)he->h_addr);
  memset(&(their_addr.sin_zero), '\0', 8); // zero the rest of the struct
  
}


  int addr_len, numbytes;
  char buf[65535];



  
  addr_len = sizeof(struct sockaddr);
  if ((numbytes=recvfrom(sockfd,buf, 65534, 0,
			 (struct sockaddr *)&their_addr, &addr_len)) == -1) {
    perror("recvfrom");
    exit(1);
  }
  


  printf("got packet from %s\n",inet_ntoa(their_addr.sin_addr));
  printf("packet is %d bytes long\n",numbytes);
  buf[numbytes] = '\0';
  printf("packet contains \"%s\"\n",buf);
  
  close(sockfd);



//////////////////////////////////////////////

  struct hostent    *host_ent;
  struct in_addr    *addr_ptr;
  struct sockaddr_in cli_addr;
  int           socket_desc, val;
  Socket        m_socket;

  m_socket.socketfd = -1;  //should change later when creating socket, otherwise error 

  // Adresse des Servers besorgen (anhand des Namens)

  // geht auch, wenn host schon ip-Adr ist?
  if((host_ent = (struct hostent *)gethostbyname(host)) == NULL) {
    // Check if a numeric address
    if((int)inet_addr(host) == -1){
      exit(-1);
    }
  }
  else{
    addr_ptr = (struct in_addr *) *host_ent->h_addr_list;
    host = inet_ntoa(*addr_ptr);
  }

  sigset_t sigmask;
  sigemptyset(&sigmask);
  sigaddset(&sigmask, SIGIO);
  sigprocmask(SIG_BLOCK, &sigmask, NULL);

  // Open UDP socket.
  if( (socket_desc = socket(AF_INET, SOCK_DGRAM, 0)) < 0){
    return m_socket;   // Can't open socket. 
  }

  //  Bind any local address.
  bzero((char *) &cli_addr, sizeof(cli_addr));
  cli_addr.sin_family       = AF_INET;
  cli_addr.sin_addr.s_addr  = htonl(INADDR_ANY);
  cli_addr.sin_port     = htons(0);

  if (fcntl(socket_desc, F_SETOWN, getpid()) == -1) return m_socket;
  val = fcntl(socket_desc, F_GETFL, 0);
  val |= O_NDELAY;
  val |= FASYNC;
  fcntl(socket_desc, F_SETFL, val) ;

  if(bind(socket_desc, (struct sockaddr *) &cli_addr,
      sizeof(cli_addr)) < 0){
    return m_socket;  // Can't bind local address 
  }



   // Fill in the structure with the address of the server.
  m_socket.socketfd = socket_desc;

  bzero((char *) &m_socket.serv_addr, sizeof(m_socket.serv_addr)) ;
  m_socket.serv_addr.sin_family     = AF_INET;
  m_socket.serv_addr.sin_addr.s_addr    = inet_addr(host);
  m_socket.serv_addr.sin_port       = htons(port);

return m_socket;
}


////////////////////////////////


bool Coach::init( const char *serverHost, int serverPort, const char *teamname ) {
  // open socket
  
  // UDP-Socket erzeugen
  int udpSocketDesc = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  
  // Socket an IP:Port binden
  if( bind(udpSocketDesc, 
	   (struct sockaddr*)&socketlist[i].sockAddrIn, 
	   sizeof(socketlist[i].sockAddrIn)) 
      != 0) {
    cerr << "bind failed" << endl;
  }
  
  


  // send init message 
  // read "init ok" from server
  // read "server_param" from server
  // read "player_param" messag
  // read "player_type" messages
  return true;
}
*/
