/******************************************************************************
 * Ball.hh
 * Copyright (c) 2003 Virtual Werder
 *
 *
 * implements:
 *   class Ball
 *
 * requires:
 *
 *
 * Created on: 21.01.2004 by aw
 * Maintainer: aw, mangels
 * Group:
 * Co-authors:
 *
 * $Id: Ball.hh,v 1.2 2004/05/13 16:33:46 aw Exp $
 *****************************************************************************/


#ifndef BALL_HH
#define BALL_HH

#include "common.hh"
#include "Geometry.hh"

using namespace std;

/**
 	This class reprents a ball with it position and velocity.
  @ingroup WorldModel
  @brief class Ball in group WorldModel
 */
class Ball {
private:
	Vector2D position;
  Vector2D velocity;
  
  /**
   * average position values
   */
  float avgX;
  float avgY;
  
  /**
   * number of used cycles
   */
  int avgCycles;
  
public:
	Ball();
  ~Ball();

  /**
   * updates the average X,Y position value
   */
  void updateAvgValues();
    

  /*
   * some set-methods
   */
	void setPosition (const Vector2D pos);
  void setPosition (const float posX, const float posY);
  void setVelocity (const Vector2D vel);
  void setVelocity (const float velX, const float velY);
  
  /**
   * calculates the average x position
   * @return float avgX
   */
  float getAvgX(void) const {
    if (avgCycles == 0) return 0;
    return (float) (avgX / avgCycles);
  }
  
  /**
   * calculates the average y position
   * @return float avgy
   */
  float getAvgY(void) const {
    if (avgCycles == 0) return 0;
    return (float) (avgY / avgCycles);
  }	
  
  Vector2D getPosition(void) const { return position; }
  float getPosX(void) const { return position.getX(); }
  float getPosY(void) const { return position.getY(); }
  float getVelX(void) const { return velocity.getX(); }
  float getVelY(void) const { return velocity.getY(); }
};

#endif // BALL_HH
