/***********************************************************************
 * Formations.cc
 * Copyright (c) 2003 Virtual Werder
 *
 * implements:
 *   class Formation
 *
 * requires:
 *
 *
 * Created on: 19.09.2003 by harlekin
 * Maintainer: cniehaus, harlekin
 * Group:
 * Co-authors: cniehaus
 *
 * Changelog:
 *
 * $Id: Formations.cc,v 1.1 2004/05/13 16:54:47 aw Exp $
 *
 **********************************************************************/

#include "Formation.hh"
#include "Formations.hh" 
#include <vector>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <fstream>

vector<Formation> Formations::formations;
Formation *Formations::currentFormation;

Formations::Formations(){
};

Formation *Formations::getFormation(int id){
  vector<Formation>::iterator i;
  for(i = formations.begin(); i!=formations.end(); ++i){
    if(i->getId() == id)
      return &(*i);
  }
#if defined(DEBUG3_HARLEKIN)
  DEBUG3_HARLEKIN("Formations::getFormation: Formation with id ", id, " not found...");
#endif
  return (&(formations.front()));
}

bool Formations::readFormations(const char *strFile){
    ifstream in( strFile );
    if( !in ){
      cerr << "Formations::readFormations: Could not open file '" <<
	strFile << "'" << endl;
      return false;
    }

    char strLine[256], *str;
    int  iLineNr          = 0, i;
    int  iLineInFormation = 0; // current offset of line in formation
    bool bReturn          = true;
    //PlayerTypeInfo *pt_info;

    int id = 0;
    double hPX[11];
    double hPY[11];
    double iPX[11];
    double iPY[11];
    int bId[11];
    double aX[11];
    double aY[11];
    double maxX[11];
    double minX[11];
    bool   bB[11];

    // read all lines
    while( bReturn && in.getline( strLine, sizeof(strLine) ) ){
      str = &strLine[0];
      iLineNr++;
      
      // comment and empty lines should be skipped
      if( !(strLine[0] == '\n' || strLine[0] == '#' || strLine[0]=='\0' ||
	    gotoFirstNonSpace( &str ) == '\0' ) ){
	// there are eleven different lines in a formation (see comment above)
	// all values for each line are parsed in one iteration
	// after all 11 lines are parsed, the sequence it is resetted.
	switch( iLineInFormation ){
        case 0: // first line is the number of the formation
          id = parseFirstInt( &str );
          break;
        case 1: // the x coordinate of the home pos for all the players
          for( i = 0 ; i < 11 ; i ++ ){
            hPX[i] = (parseFirstDouble(&str));
	  }
          break;
        case 2: // the y coordinate of the home pos for all the players
          for( i = 0 ; i < 11 ; i ++ ){
            hPY[i] = (parseFirstDouble(&str));
	  }
          break;
        case 3: // the x coordinate of the init pos for all the players
          for( i = 0 ; i < 11 ; i ++ ){
            iPX[i] = (parseFirstDouble(&str));
	  }
          break;
        case 4: // the y coordinate of the init pos for all the players
          for( i = 0 ; i < 11 ; i ++ )
            iPY[i] = (parseFirstDouble(&str));
          break;
        case 5: // the player behaviour for all players
          for( i = 0 ; i < 11 ; i ++ )
            bId[i] = parseFirstInt(&str);
          break;
        case 6: // the x attraction for all the player types
          for( i = 0 ; i < 11 ; i ++ )
	    aX[i] = parseFirstDouble( &str );
          break;
        case 7: // the y attraction for all the player types
          for( i = 0 ; i < 11 ; i ++ )
	    aY[i] = parseFirstDouble( &str );
          break;
        case 8: // stay behind the ball for all the player types
          for( i = 0 ; i < 11 ; i ++ ){
            if( parseFirstInt( &str ) == 1 )
              bB[i] = true;
            else
              bB[i] = false;
          }
          break;
        case 9: // the minimal x coordinate for all the player types
          for( i = 0 ; i < 11 ; i ++ )
	    minX[i] = parseFirstDouble( &str );
          break;
        case 10:// the maximal x coordinate for all the player types
          for( i = 0 ; i < 11 ; i ++ )
	    maxX[i] = parseFirstDouble( &str );
          break;
        default:
          cerr << "Formations::readFormations: error line " << iLineNr <<endl;
          return false;
	}
      

	iLineInFormation++;         // go one line further
	if( iLineInFormation == 11 ){ // each formation consists of eleven lines

	  iLineInFormation = 0;
	  //DEBUG2_HARLEKIN("Formations: Read Formation with ID: ", id);
	  Formation f = Formation(id, bId, hPX, hPY, iPX, iPY, aX, aY, maxX, minX, bB);
	  addFormation( f );
	}
      }
    }
#if defined(DEBUG3_HARLEKIN)
    DEBUG3_HARLEKIN("Formations: Read ", formations.size(), " formations");
#endif
    return true;
}

char Formations::gotoFirstNonSpace( char** strMsg ){
  while(*strMsg && isspace(**strMsg) )
    (*strMsg)++;
  return  (*strMsg) ?  **strMsg : '\0';
}

int Formations::parseFirstInt( char** strMsg ){
  int  iRes  = 0;
  bool bIsMin= false;
  char *str  = *strMsg;
  
  while( *str != '\0' && !isdigit(*str) && *str!='-')
    str++;                             // walk to first non digit or minus sign
  if( *str != '\0' && *str=='-'){
    bIsMin=true;
    str++;
  }
  while( *str != '\0' && *str<='9' && *str>='0' ){
      iRes=iRes*10+(int)(*str-'0');                 // multiply old res with 10
      str++;                                        // and add new value
  }
  *strMsg = str;
  return (bIsMin) ? -iRes : iRes;
}

int Formations::createFormationFrom(int defender, int attacker){
  int midfielder = 10 - defender - attacker;
  int highestID = 0;
  vector<Formation>::iterator it_f;
  for (it_f = formations.begin(); it_f != formations.end(); it_f++){
    if (it_f->id > highestID){
      highestID = it_f->id;
    }
  }
  Formation f = Formation();
  f.id = highestID + 1;

  f.behaviour[0]  = B_Goalie;
  f.homePos[0]    = Vector2D(-50.0,   0.0);
  f.initPos[0]    = Vector2D(-45.0,   0.0);
  f.attrX[0]      = 0.1;
  f.attrY[0]      = 0.1;
  f.maxX[0]       = -52.5;
  f.minX[0]       = -30.0;
  f.behindBall[0] = true;
  
  int cUnum = 2;
  for (int i = 0; i < defender; i++){
    f.behaviour[cUnum-1]  = B_CenterBackDefenderLeft;
    f.homePos[cUnum-1]    = Vector2D(-18.0,   0.0); // y will be set later
    f.initPos[cUnum-1]    = Vector2D(-16.0,   0.0);
    f.attrX[cUnum-1]      = 0.50;
    f.attrY[cUnum-1]      = 0.25;
    f.maxX[cUnum-1]       = -10.0;
    f.minX[cUnum-1]       = -41.0;
    f.behindBall[cUnum-1] = true;
    cUnum++;
  }
  for (int i = 0; i < midfielder; i++){
    f.behaviour[cUnum-1]  = B_CenterMidfielder;
    f.homePos[cUnum-1]    = Vector2D(  0.0,   0.0); // y will be set later
    f.initPos[cUnum-1]    = Vector2D(-10.0,   0.0);
    f.attrX[cUnum-1]      = 0.50;
    f.attrY[cUnum-1]      = 0.25;
    f.maxX[cUnum-1]       =  35.0;
    f.minX[cUnum-1]       = -38.0;
    f.behindBall[cUnum-1] = false;
    cUnum++;
  }
  for (int i = 0; i < attacker; i++){
    f.behaviour[cUnum-1]  = B_CenterAttacker;
    f.homePos[cUnum-1]    = Vector2D( 20.0,   0.0); // y will be set later
    f.initPos[cUnum-1]    = Vector2D( -1.0,   0.0);
    f.attrX[cUnum-1]      = 0.60;
    f.attrY[cUnum-1]      = 0.25;
    f.maxX[cUnum-1]       =  42.0;
    f.minX[cUnum-1]       =   0.0;
    f.behindBall[cUnum-1] = false;
    cUnum++;
  }
  cUnum = 2;
  distributePlayers(&f, 1, defender,   2);
  distributePlayers(&f, 2, midfielder, 2 + defender);
  distributePlayers(&f, 3, attacker,   2 + defender + midfielder);
  f.initPos[11-1] = Vector2D(0.0, 0.0);
  formations.push_back(f);
  f.print();
  return f.id;
}

void Formations::distributePlayers(Formation *f, int type, int amount, int start){  
  int tL, tCL, tCR, tR;
  switch(type){
  case 1:
    tL  = B_WingDefenderLeft;
    tCL = B_CenterBackDefenderLeft;
    tCR = B_CenterBackDefenderRight;
    tR  = B_WingDefenderRight;
    break;
  case 2:
    tL  = B_WingMidfielderLeft;
    tCL = tCR = B_CenterMidfielder;
    tR  = B_WingMidfielderRight;
    break;
  case 3:
    tL  = B_WingAttackerLeft;
    tCL = tCR = B_CenterAttacker;
    tR  = B_WingAttackerRight;
    break;
  default:
    return;
  }

  bool sym = !(amount % 2);
  int leftFrom = start;
  int leftTo;
  int rightTo = start + amount - 1;
  float ylFrom = 20;
  float step = ylFrom * 2 / amount; 
  //float ylTo;
  if (!sym) {
    leftTo = leftFrom - 1 + (int)floor(((float)amount)/2.0);
  }else{
    leftTo = leftFrom - 1 + amount/2;
  }
  int count = leftTo - leftFrom;

  for (int i = 0; i <= count; i++){
    int unumL = leftFrom + i;
    int unumR = rightTo  - i;
    if (sym && amount > 2 && (i + 1) == leftTo){      
      f->behaviour[unumL-1] = tCL;
      f->behaviour[unumR-1] = tCR;
    }else{
      f->behaviour[unumL-1] = tL;
      f->behaviour[unumR-1] = tR;
    }
    float yVal = (ylFrom - step * i);
    f->homePos[unumL-1] = Vector2D(f->homePos[unumL-1].getX(), -yVal);
    f->homePos[unumR-1] = Vector2D(f->homePos[unumR-1].getX(),  yVal);
    f->initPos[unumL-1] = Vector2D(f->initPos[unumL-1].getX(), -yVal * 0.75);
    f->initPos[unumR-1] = Vector2D(f->initPos[unumR-1].getX(),  yVal * 0.75);    
  }

}

double Formations::parseFirstDouble( char** strMsg ){
  double dRes=0.0, dFrac=1.0;
  bool bIsMin=false, bInDecimal=false, bCont=true;
  char *str = *strMsg;

  // go to first part of double (digit, minus sign or '.')
  while( *str != '\0' && !isdigit(*str) && *str!='-' && *str!='.'){
    // when NaN or nan is double value, return -1000.0
    if( (*str=='N' && strlen(str)>3 && *(str+1)=='a' && *(str+2)=='N') ||
        (*str=='n' && strlen(str)>3 && *(str+1)=='a' && *(str+2)=='n') ){
      *strMsg = str+3;
      return -1000.0;
    }
    else
       str++;
  }

  if( *str != '\0' && *str=='-'){
    bIsMin=true;
    str++;
  }

  while( *str != '\0' && bCont) {
    if( *str=='.' )                            // in decimal part after '.'
      bInDecimal = true;
    else if( bInDecimal && *str<='9' && *str>='0') // number and in decimal
    {
      dFrac=dFrac*10.0;                        // shift decimal part to right
      dRes += (double)(*str-'0')/dFrac;        // and add number
    }
    else if( *str<='9' && *str>='0' )          // if number and not decimal
      dRes=dRes*10+(double)(*str-'0');         // shift left and add number
    else if( *str=='e' ) // 10.6e-08           // if to power e
    {
      if( *(str+1) == '+' )                    // determine sign
        dRes *=  pow((double)10, parseFirstInt(&str)); // and multiply with power
      else if( *(str+1) == '-' )
      {
        str = str+2;                           // skip -
        dRes /=  pow((double)10, parseFirstInt(&str)); // and divide by power
      }
      bCont = false;                           // after 'e' stop
    }
    else
      bCont = false;                           // all other cases stop

    if( bCont == true )                        // update when not stopped yet
      str++;
  }
  *strMsg = str;
  return (bIsMin && dRes != 0.0) ? -dRes : dRes;
}
