/******************************************************************************
 * PlayerType.hh
 * Copyright (c) 2003 Virtual Werder
 *
 *
 * implements:
 *   class PlayerType
 *
 * requires:
 *
 *
 * Created on: 15.09.2003 by harlekin
 * Maintainer: cniehaus, harlekin
 * Group:
 * Co-authors: cniehaus
 *****************************************************************************/

#ifndef PLAYERTYPE_HH
#define PLAYERTYPE_HH

#include "common.hh"


/**
	This class represents a player type. The default type's id is 0. Heterogenous
	types have ids from 1 to 6. They can be distributed by the coach, to take advantage
	of the different player types.
	@ingroup WorldModel
  @brief class PlayerType in group WorldModel
 */
class PlayerType {
public:
  PlayerType () {};
  virtual ~PlayerType () {};

  void setAll(
	      const int id,
	      const float player_speed_max,
	      const float stamina_inc_max,
	      const float player_decay,
	      const float inertia_moment,
	      const float dash_power_rate,
	      const float player_size,
	      const float kickable_margin,
	      const float kick_rand,
	      const float extra_stamina,
	      const float effort_max,
	      const float effort_min
	      ){
    this->id       = id;
    playerSpeedMax = player_speed_max;
    staminaIncMax  = stamina_inc_max;
    playerDecay    = player_decay;
    inertiaMoment  = inertia_moment;
    dashPowerRate  = dash_power_rate;
    playerSize     = player_size;
    kickableMargin = kickable_margin;
    kickRand       = kick_rand;
    extraStamina   = extra_stamina;
    effortMax      = effort_max;
    effortMin      = effort_min;  
  };

  int getTypeId () const { return id; }
  float getPlayerSpeedMax () const { return playerSpeedMax; }
  float getStaminaIncMax () const { return staminaIncMax; }
  float getPlayerDecay () const { return playerDecay; }
  float getInertiaMoment () const { return inertiaMoment; }
  float getDashPowerRate () const { return dashPowerRate; }
  float getPlayerSize () const { return playerSize; }
  float getKickableMargin () const { return kickableMargin; }
  float getKickRand () const { return kickRand; }
  float getExtraStamina() const { return extraStamina; }
  float getEffortMax () const { return effortMax; }
  float getEffortMin () const { return effortMin; }

  float aValue, mValue, dValue;
  unsigned int used;

private:
  int id;
  float playerSpeedMax;
  float staminaIncMax;
  float playerDecay;
  float inertiaMoment;
  float dashPowerRate;
  float playerSize;
  float kickableMargin;
  float kickRand;
  float extraStamina;
  float effortMax;
  float effortMin;
};

#endif // PLAYERTYPE_HH
