/******************************************************************************
 * WorldModel.cc
 * Copyright (c) 2003 Virtual Werder
 *
 *
 * implements:
 *   class WorldModel
 *
 * requires:
 *
 *
 * Created on: 21.01.2004 by aw
 * Maintainer: aw, mangels
 * Group:
 * Co-authors:
 *
 * $Id: WorldModel.cc,v 1.10 2004/06/29 09:12:39 aw Exp $
 *****************************************************************************/

#include "WorldModel.hh"
#include "debug.hh"
#include "common.hh"
#include "ConfigFile.hh"


#define PLAYER_SIZE 0.3
#define BALL_SIZE 0.085

// how often balancing can be called from itself...
#define BALANCE_COUNT_MAX 25

WorldModel::WorldModel() {
  ptPlayerSpeedMax    = 0;
  ptKickableMarginMax = 0;
  balanceCount = 0;
  otherDefenseLineX = 0;
  ourFormation = ConfigFile::standardFormation;
  otherFormation = FO_Unknown;
  score = 0;
}


WorldModel::~WorldModel(){

}


/*
 * Integrates the information provided by the 'SeeGlobalMsg'
 */
void WorldModel::integrateSeeGlobalMsg(SeeGlobalMsg *msg)
{
  time = msg->serverTime;

  bool playerExists = false;
  /*
   * Update myTeam
   */
  vector<ParsedPlayer *>::const_iterator i;
  for (i = (*(msg->ourTeam)).begin(); i!= (*(msg->ourTeam)).end(); ++i) {
    playerExists = false;
#if defined (DEBUG_WORLDMODEL)
    cout << "Updating player (" << (*i)->uNum << ") in 'myTeam'!" << endl;
#endif
    vector<Player>::iterator j;
    for (j = myTeam.begin(); j!= myTeam.end() && playerExists == false; ++j) {
      if ((*i)->uNum == j->getUNum()) {
        // UPDATE PLAYER
	playerExists = true;
	j->setPosition((*i)->posX, (*i)->posY);
	j->setVelocity((*i)->velX, (*i)->velY);
	j->setBodyDir((*i)->bodyAngle);
	j->setNeckDir((*i)->neckAngle);
	j->setGoalie((*i)->goalie);
      }
    }
    if (!playerExists) {
      // ADD PLAYER
      Player p((*i)->uNum);
      p.setPosition((*i)->posX, (*i)->posY);
      p.setVelocity((*i)->velX, (*i)->velY);
      p.setBodyDir((*i)->bodyAngle);
      p.setNeckDir((*i)->neckAngle);
      p.setGoalie((*i)->goalie);
      myTeam.push_back(p);
    }
  }

  /*
   * Update otherTeam
   */
  playerExists = false;
  //vector<ParsedPlayer *>::const_iterator i;
  for (i = (*(msg->theirTeam)).begin(); i!= (*(msg->theirTeam)).end(); ++i) {
    playerExists = false;
#if defined (DEBUG_WORLDMODEL)
    cout << "Updating player (" << (*i)->uNum << ") in 'otherTeam' !" << endl;
#endif
    vector<Player>::iterator j;
    for (j = otherTeam.begin(); j!= otherTeam.end() && playerExists == false; ++j) {
      if ((*i)->uNum == j->getUNum()) {
        // UPDATE PLAYER
	playerExists = true;
	j->setPosition((*i)->posX, (*i)->posY);
	j->setVelocity((*i)->velX, (*i)->velY);
	j->setBodyDir((*i)->bodyAngle);
	j->setNeckDir((*i)->neckAngle);
	j->setGoalie((*i)->goalie);
      }
    }
    if (!playerExists) {
      // ADD PLAYER
      Player p((*i)->uNum);
      p.setPosition((*i)->posX, (*i)->posY);
      p.setVelocity((*i)->velX, (*i)->velY);
      p.setBodyDir((*i)->bodyAngle);
      p.setNeckDir((*i)->neckAngle);
      p.setGoalie((*i)->goalie);
      otherTeam.push_back(p);
    }
  }

  /*
   * Update the ball
   */
  ball.setPosition(msg->ball->posX, msg->ball->posY);
  ball.setVelocity(msg->ball->velX, msg->ball->velY);
}


void WorldModel::integrateChangePlayerTypeMsg (ChangePlayerTypeMsg *msg) {
  //time = msg->serverTime;
  printf  ("otherTeam: Type of player %i changed!\n", msg->uNum);
  vector<Player>::iterator j;
  for (j = otherTeam.begin(); j!= otherTeam.end(); ++j) {
    if (msg->uNum == j->getUNum()) {
      // UPDATE PLAYER
      j->setHasTypeChanged(true);
      if (time > 0) 
        j->resetPlayerTypePrediction();
      return;
    }
  }
  // Player not found: New Player....
  Player p(msg->uNum);
  p.setHasTypeChanged(true);
  otherTeam.push_back(p);
}


void WorldModel::determineHeterogenousPlayers(){
  // get fastest playertype
  float ptMaxSpeeds[7] = {0};
  //int fastestPt = 0;
  if (!(*playerTypes).empty()) {
    vector<PlayerType>::const_iterator pt;
    int idx = 0;
    //float lastSpeed = 0;
    for (pt = (*playerTypes).begin(); pt != (*playerTypes).end(); ++pt){
      ptMaxSpeeds[idx] = (pt->getDashPowerRate() * pt->getEffortMax() * 100) / (1 - pt->getPlayerDecay());
#if defined (DEBUG_DETERMINE_HETEROGENOUS)
      printf("  | MaxSpeed (without decay) for PlayerType %i, speed: %.3f |\n", idx, ptMaxSpeeds[idx]); 
#endif
      idx++;
    }
    printf("  `---------------------------------------------------------'\n");
  }
  
  vector<Player>::iterator i;	
  // find most fitting player types
  for (i = otherTeam.begin(); i != otherTeam.end(); ++i) {
    // skipping player 1 (the goalie .. is that a good idea? 'cause the goalie can have another unum than 1!) 
    // and skipping players without type changes
    //if (i->getUNum() == 1 || !i->hasTypeChanged()) {
    if (!i->hasTypeChanged()) {
#if defined (DEBUG_DETERMINE_HETEROGENOUS)      
      printf ("determineHeterogenousPlayers(): Skipping player %i\n", i->getUNum());
#endif
      continue;
    }
    
    /*   That will skip further calculations, for a given player... if we have found a playertype... !?
    if (i->playerType != -1) {
      printf("###Player %i: playerType %i\n", i->getUNum(), i->playerType);
      continue;
    }
    */
    
    if (!(*playerTypes).empty()) {
      vector<PlayerType>::const_iterator pt;
      int idx = 0;
      int typeId = 0;
      float matching = 0;
      for (pt = (*playerTypes).begin(); pt != (*playerTypes).end(); ++pt) {
        typeId = pt->getTypeId();
        
        float maxSpeed = ((/*min(1.2f,*/ptMaxSpeeds[idx]) * (pt->getPlayerDecay()));
        float maxAccel = 100.0f * pt->getDashPowerRate() * pt->getEffortMax();

        matching = i->getPlayerTypeMatching(typeId,
            maxSpeed,
            pt->getKickableMargin(),
            pt->getInertiaMoment(),
            maxAccel,
	    pt->getPlayerDecay());
        
#if defined (DEBUG_DETERMINE_HETEROGENOUS)        
        if (idx == 0)
          printf ("   * Values for player %i at cycle %i (pVMax %.3f, kMargin %.3f, accel %.3f, inertia %.3f)\n", 
              i->getUNum(), 
              time,
              i->getTopSpeed(), 
              i->getKickableMargin(),
              i->getAccel(),
              i->getInertia());
          
        printf ("\t@type %i matching %03.0f (ptVMax %.3f, Vdiff %.3f)\n", 
            idx,
            matching,
            maxSpeed, 
            fabs(i->getTopSpeed() - maxSpeed));
        
        if (idx == 6)
          printf ("\n");
#endif
        idx++;
      }
    }

/*    
#if defined (DEBUG_DETERMINE_HETEROGENOUS)
    printf("possiblePlayerTypes for player %i at %i:\n", i->getUNum(), time);
    printf("  values, speedMax: %.4f, kickMargin: %.4f\n", i->getTopSpeed(), i->getKickableMargin());
    for (int j = 0; j < 7; j++){
      printf("  playerType %i: %.4f\n", j, (i->getPlayerTypePossible(j)));
    }
    if (i->getUNum() == 11) printf ("\n\n");
#endif
*/
    //i->checkIfDoneWithPlayerTypes();
  }
  // checkDuplicatePlayerTypes();
  
  // this causes a recursive method call.. be carefull
  if (balancePlayerTypeDetermination())
    printf ("Ok, all player types seem to be correctly balanced! Printing prediction...\n");
  else
    printf ("PlayerTypes are not balanced yet! Printing prediction...\n");
  
  printPlayerTypePrediction();
}

void WorldModel::calcPlayerSpeedMax(){
  if (wmstates.size() < 3) return;
	
  // last worldmodelstate = current worldmodel!
  // jump to previous state
  vector<WorldModelState>::iterator i;
  i = wmstates.end();
  i -= 2;
  
#if defined (DEBUG_PLAYER_SPEED_MAX)
  printf ("calcPlayerSpeedMax(): TopSpeed %i/%i Updating..", time, i->time);
#endif
  float pspeed = 0;
  vector<Player>::iterator n;
  for (n = i->otherTeam.begin(); n != i->otherTeam.end(); ++n) {
    vector<Player>::iterator j;
    for (j = otherTeam.begin(); j != otherTeam.end(); ++j) {
      if (j->getUNum() == n->getUNum() && j->getVelocity() >= n->getTopSpeed() ) {
	pspeed = j->getTopSpeed();
	j->setTopSpeed(j->getVelocity());
#if defined (DEBUG_PLAYER_SPEED_MAX)
	printf ("(%i %f (%f)) ", j->getUNum(), j->getTopSpeed(), pspeed);
#endif
      }
    }
  }
#if defined (DEBUG_PLAYER_SPEED_MAX)
  printf ("\n");
#endif
}


void WorldModel::calcKickableMargin(){
	
  if (wmstates.size() < 4) return;
	
  if (ptKickableMarginMax == 0) 
    getPTMaxValues();
	
  vector<WorldModelState>::iterator i;
  i = wmstates.end();
  // last worldmodelstate (i -1) == current worldmodel!
  i -= 2;
  if (i->playmode != PM_PlayOn) return;
  Vector2D p1 = (i->ball).getPosition();
  i--;
  if (i->playmode != PM_PlayOn) return;
  Vector2D p2 = (i->ball).getPosition();
  i--;
  if (i->playmode != PM_PlayOn) return;
  Vector2D p3 = (i->ball).getPosition();
	
  float l1, l2;
	
  // speed of ball
  l1 = p2.distanceTo(p3); // cycle t-2
  l2 = p1.distanceTo(p2); // cycle t-1

  bool kicked = false;

  float diff = l1 - l2;
  // ball is faster than he was last cycle
  if (diff < 0) 
    kicked = true;
  // ball is slower or faster than possible 
  // by balldecay ([0.89..0.99])
  // use some tolerance (0.1)
  if ((l1 * 0.99) - l2 < -0.1 ||
      (l1 * 0.89) - l2 >  0.1 )
    kicked = true;
  
  if (kicked){
#if defined (DEBUG_KICKABLE_MARGIN)	
    printf ("calcKickableMargin(): time %d ball changed direction!\n", time);
#endif
	
    // increment i to the current WorldModelState, with ball position p2
    i++;
    	
    // temporary distance value, used for calculation
    float dist;
    	
    // distance of nearest player of my team to the ball
    float mTMinDist = -1;
    // unum of this player
    int mTUNum = -1;
    vector<Player>::iterator j;
    for (j = i->myTeam.begin(); j != i->myTeam.end(); ++j) {
      dist = p2.distanceTo(j->getPosition()) - PLAYER_SIZE - BALL_SIZE; 
      if (dist < mTMinDist || mTMinDist < 0) {
	mTMinDist = dist;
	mTUNum = j->getUNum();
      } 
    }
    	
    // distance of nearest player of other team to the ball
    float oTMinDist = -1;
    // unum of this player
    int oTUNum = -1;
    for (j = i->otherTeam.begin(); j != i->otherTeam.end(); ++j) {
      dist = p2.distanceTo(j->getPosition()) - PLAYER_SIZE - BALL_SIZE; 
      if (dist < oTMinDist || oTMinDist < 0) {
	oTMinDist = dist;
	oTUNum = j->getUNum();
      }
    }
    
    if (oTMinDist < mTMinDist) {
      // OTHER TEAM
      for (j = otherTeam.begin(); j != otherTeam.end(); ++j) {
	if (oTMinDist < ptKickableMarginMax && 
	    mTMinDist > ptKickableMarginMax &&
	    j->getUNum() == oTUNum && 
	    j->getKickableMargin() < oTMinDist) {
	  j->setKickableMargin(oTMinDist);
	}
      }
    }
  }
}


void WorldModel::calcAcceleration(){
  if (wmstates.size() < 4) return;
  vector<WorldModelState>::iterator i;
  i = wmstates.end();
  i -= 4;
  if (i->playmode != PM_PlayOn) return;
  Vector2D predPosition1[11];
  Vector2D predPosition2[11];
  vector<Player>::iterator j;
  for (j = i->otherTeam.begin(); j != i->otherTeam.end(); ++j) {
    predPosition2[j->getUNum()-1] = j->getPosition(); 
  }
  i++;
  if (i->playmode != PM_PlayOn) return;
  for (j = i->otherTeam.begin(); j != i->otherTeam.end(); ++j) {
    predPosition1[j->getUNum()-1] = j->getPosition(); 
  }
  i++;
  if (i->playmode != PM_PlayOn) return;

  for (j = i->otherTeam.begin(); j != i->otherTeam.end(); ++j) {
    int unum = j->getUNum();
    float d1, d2;
    d1 = predPosition2[unum-1].distanceTo(predPosition1[unum-1]);
    d2 = predPosition1[unum-1].distanceTo(j->getPosition());

    float accel = d2 - (d1*0.7);
    if (accel <= 0) continue;
    if (accel >= 2) continue;

    vector<Player>::iterator k;
    for (k = otherTeam.begin(); k != otherTeam.end(); k++){
      if (unum != k->getUNum()) continue;
      if (accel > k->getAccel() || k->getAccel() < 0){
	k->setAccel(accel);
      }
    }    
  }
}

void WorldModel::calcInertiaMoment() {
  bool debug = false;
#if defined (DEBUG_CALCINERTIAMOMENT)
  debug = true;
#endif

  if (wmstates.size() < 2) return;
	
  vector<WorldModelState>::iterator i;
  i = wmstates.end();
  i -= 3;
  if (i->playmode != PM_PlayOn) return;
	
  //int prevtime = i->time;
  float prevBodyDir[11] = {0};
  float prevVelocity[11] = {0};
  vector<Player>::iterator j;
  for (j = i->otherTeam.begin(); j != i->otherTeam.end(); ++j) {
    prevBodyDir[j->getUNum()-1]  = j->getBodyDir();
    prevVelocity[j->getUNum()-1] = j->getVelocity();   
  }
	
  i++;
  if (i->playmode != PM_PlayOn) return;
  //vector<Player>::iterator j;
  float speed   = 0;
  float angle   = 0;
  //float inertia = 0;
  //float moment  = 0;
  for (j = i->otherTeam.begin(); j != i->otherTeam.end(); ++j) {
    //actual_angle = moment/(1.0 + inertia moment * player speed)

    int unum = j->getUNum();    
    speed = prevVelocity[j->getUNum()-1];
    angle = fabs(normalizeTo180Deg(j->getBodyDir() - prevBodyDir[j->getUNum()-1]));
    
    if (angle == 0.0) {
      if (debug) printf ("calcInertiaMoment(): continueing due to angle == 0.0\n");
      continue;
    }
    if (speed <= 0.5) {
      if (debug) printf ("calcInertiaMoment(): continueing due to speed <= 0.5\n");
      continue;
    }
    float inertia = ((180.0/angle)-1.0)/speed;
    if (debug) printf("Inertia from angle %.2f, speed %.2f: %.2f\n", angle, speed, inertia);
    
    vector<Player>::iterator k;
    for (k = otherTeam.begin(); k != otherTeam.end(); k++){
      if (unum != k->getUNum()) continue;
      if (inertia < k->getInertia() || k->getInertia() < 0){
	k->setInertia(inertia);
	continue;
      }
    }
  }
}


void WorldModel::calcDecay() {
  if (wmstates.size() < 2) return;

  vector<WorldModelState>::iterator i;
  i = wmstates.end();
  i -= 2;
  if (i->playmode != PM_PlayOn) return;
	
  //int prevtime = i->time;
  float prevBodyDir[11] = {0};
  float prevVelocity[11] = {0};
  vector<Player>::iterator j;
  for (j = i->otherTeam.begin(); j != i->otherTeam.end(); ++j) {
    prevBodyDir[j->getUNum()-1]  = j->getBodyDir();
    prevVelocity[j->getUNum()-1] = j->getVelocity();   
  }
	
  i++;
  if (i->playmode != PM_PlayOn) return;
  //vector<Player>::iterator j;
  float speed   = 0;
  float angle   = 0;
  //float inertia = 0;
  //float moment  = 0;
  for (j = i->otherTeam.begin(); j != i->otherTeam.end(); ++j) {
    speed = prevVelocity[j->getUNum()-1];
    angle = fabs(normalizeTo180Deg(j->getBodyDir() - prevBodyDir[j->getUNum()-1]));
    if (angle == 0.0) continue;
    if (speed <= 0.2) continue;

    float decay = j->getVelocity() / speed;

    vector<Player>::iterator k;
    for (k = otherTeam.begin(); k != otherTeam.end(); k++){ 
	if (j->getUNum() == k->getUNum()){
	    float decayNum = k->getDecayNum();
	    if(k->getDecay() == -1)
		k->setDecay(decay);
	    float allDecay = (decay + (decayNum * k->getDecay())) / (1 + decayNum);
#if defined (DEBUG_DECAY)
	    printf("update decay with %.2f : now average %.2f with values %.1f by player %d\n",decay, allDecay, decayNum, k->getUNum());
#endif 
	    k->setDecay(allDecay);
	}
    }
  }
}

void WorldModel::calcOtherDefenseLineX() {
  if (wmstates.size() < 2) return;
  vector<WorldModelState>::iterator i;
  i = wmstates.end();
  i -= 2;
  if (i->playmode != PM_PlayOn) return;

  bool debug = false;
#if defined (DEBUG_CALC_OTHER_DEFENSE_LINE_X)
  debug = true;
#endif
  // side of the other team.. factor 1 right, factor -1 left
  //float side = 1;
  float minX = -1;
  vector<Player>::iterator p;
  if (debug) printf ("WorldModel::calcOtherDefenseLine() :\n   ");
  for (p = i->otherTeam.begin(); p != i->otherTeam.end(); ++p) {
    if (p->isGoalie()) {
      //printf ("Other Goalie @ %f Ball @ %f\n",p->getPosition().getX(), ball.getPosition().getX());
      continue;
    }
    // get current offsideX
    if ((p->getPosition().getX() > minX || minX == -1) && 
      ball.getPosition().getX() < -15) {
      minX = p->getPosition().getX();
    }
  }

  // set minimal offsideX
  
  if ((minX < this->otherDefenseLineX || 
       this->otherDefenseLineX == 0) && minX >= 0){
    this->otherDefenseLineX = minX;
  }
  
  
  if (debug) {
    printf ("\n --> MinX (@cycle %i) is %0.2f, MinX (@cycle ALL) is %0.2f\n",
	    this->time,
	    minX,
	    this->otherDefenseLineX);
  }

}


int WorldModel::updateFormationDetection () {
  bool debug = false;
#if defined (DEBUG_UPDATE_FORMATION_DETECTION)
  debug = true;
#endif

  // skip next 50 cycles
  if (playmode == PM_TheirCornerKick
      || playmode == PM_OurCornerKick
      || playmode == PM_TheirPenaltyKick
      || playmode == PM_OurPenaltyKick) return 50;
  // skip this cycle
  if (playmode != PM_PlayOn) return 0;

  if (debug) printf ("WorldModel::updateFormationDetection(): cycle %i, playmode %i\n", time, playmode);
  
  vector<Player>::iterator p;
  for (p = otherTeam.begin(); p != otherTeam.end(); ++p) {
    if (p->isGoalie()) continue;
    p->updateAvgValues();
    if (debug) printf ("\tPlayer %i, avgX %f, avgY %f\n", p->getUNum(), p->getAvgX(), p->getAvgY());
  }
  if (debug) printf ("\n");
  
  // update the average ball position
  ball.updateAvgValues();
  
  // calculate next cycle
  return 0;
}

void WorldModel::determineFormation() {
  bool debug = false;
#if defined (DEBUG_DETERMINE_FORMATION)
  debug = true;
#endif  
  float xMax, xMin, yMax, yMin, xMiddle, yMiddle = 0;
  float tmp = 0;
  int players = 0;
  vector<Player>::iterator p;
  for (p = otherTeam.begin(); p != otherTeam.end(); ++p) {
    if (p->isGoalie()) continue;
    players++;
    
    tmp = p->getAvgX();
    if (players == 1) xMax = tmp, xMin = tmp;
    else if (tmp > xMax) xMax = tmp;
    else if (tmp < xMin) xMin = tmp;
    xMiddle += tmp;
    
    tmp = p->getAvgY();
    if (players == 1) yMax = tmp, yMin = tmp;
    else if (tmp > yMax) yMax = tmp;
    else if (tmp < yMin) yMin = tmp;
    yMiddle += tmp;
  }
  
  if (players == 0) return;
  xMiddle = (float) xMiddle / players;
  yMiddle = (float) yMiddle / players;

  if (debug) printf ("WorldModel::determineFormation(): Players %i, xMax %f, xMin %f, yMax %f, yMin %f,\n\txMiddle %f, yMiddle %f, ballXMiddle %f, ballYMiddle %f, Score %i\n", players, xMax, xMin, yMax, yMin, xMiddle, yMiddle, ball.getAvgX(), ball.getAvgY(), score);


  int subX[3] = {0};
  float d1, d2, d3 = {0};
  for (p = otherTeam.begin(); p != otherTeam.end(); ++p) {
    if (p->isGoalie()) continue;
    tmp = p->getAvgX();
    d1 = fabs(tmp - xMin);
    d2 = fabs(tmp - xMiddle);
    d3 = fabs(tmp - xMax);
    
    if      (d1 < d2 && d1 < d3) subX[2]++; // add one attacker
    else if (d2 < d1 && d2 < d3) subX[1]++; // add one midfielder
    else if (d3 < d1 && d3 < d2) subX[0]++; // add one defender
  }
  
  /*
   *      GET THE BASIC FORMATION OF THE OTHER TEAM...
   */
  if      (subX[0] == 3 && subX[1] == 4 && subX[2] == 3) otherFormation = FO_343;
  else if (subX[0] == 4 && subX[1] == 4 && subX[2] == 2) otherFormation = FO_442;
  else if (subX[0] == 5 && subX[1] == 3 && subX[2] == 2) otherFormation = FO_532;
  else if (subX[0] == 4 && subX[1] == 2 && subX[2] == 4) otherFormation = FO_424;
  else if (subX[0] == 3 && subX[1] == 5 && subX[2] == 2) otherFormation = FO_352;
  else if (subX[0] == 4 && subX[1] == 3 && subX[2] == 3) otherFormation = FO_433;
  else if (subX[0] == 3 && subX[1] == 3 && subX[2] == 4) otherFormation = FO_334;
  else if (subX[0] == 3 && subX[1] == 2 && subX[2] == 5) otherFormation = FO_325;
  else otherFormation = FO_Unknown;
  
  //float fieldside = ball.getAvgX() + xMiddle;
  int fieldside = 0;
  if (ball.getAvgX() < 0 && xMiddle < 0) fieldside = -2;
  if (ball.getAvgX() < 0 && xMiddle >= 0) fieldside = -1;
  if (ball.getAvgX() > 0 && xMiddle > 0) fieldside = 2;
  if (ball.getAvgX() > 0 && xMiddle <= 0) fieldside = 1;
  
  /*
   *      FIND THE APPROPRIATE TACTIC
   */
  // other team is dominating, on our field side and has more goals
  if (fieldside < 0 && score < -1)
    ourTactic = TA_Defensive;
  
  // other team is dominating, on our field side and has ONE more goal and there are only 1000 cycles left...
  else if (fieldside < 0 && score == -1 && time >= 5000)
    ourTactic = TA_Offensive;
  
  // we are more often on their field side, but they have a better score
  else if (fieldside > 0 && score < 0)
    ourTactic = TA_Offensive;
  
  // we are more often on their field side, and the score diff > 1
  else if (fieldside > 0 && score > 1)
    ourTactic = TA_Offensive;
  
  // we are more often on their field side, and the score diff == 1 and there are only 1000 cycles left
  else if (fieldside > 0 && score == 1 && time >= 5000)
    ourTactic = TA_Defensive;
  
    // other team is more often on our field side, and the score diff == 1 and there are only 1000 cycles left
  else if (fieldside < 0 && score == 1 && time >= 5000)
    ourTactic = TA_Defensive;
  
  // we are more often on their field side, and the score is equal and more than 4000 cycles have been passed
  else if (fieldside > 1 && score == 0 && time > 4000)
    ourTactic = TA_Offensive;
  
  // else use default tactic
  else 
    ourTactic = TA_Default;
  
  
  /*
   *      FIND THE BEST FORMATION FOR OUR TEAM
   */ 
  if (debug) printf ("WorldModel::determineFormation(): Defenders %i, Midfielders %i, Attackers %i\n", subX[0], subX[1], subX[2]);
  switch (otherFormation) {
    case FO_343:
      if (debug) printf ("Other formation is FO_343\n");
      if (ourTactic == TA_Defensive)
        setFormation(ConfigFile::fo_442_def);               // 442 Defensiv
      else if (ourTactic == TA_Offensive)
        setFormation(ConfigFile::fo_442_off);               // 442 Angriff
      else 
        setFormation(ConfigFile::fo_442);                   // 442 GO2004 <Die Raute> German-Open
      break;
      
    case FO_442:
      if (debug) printf ("Other formation is FO_442\n");
      if (ourTactic == TA_Defensive)
        setFormation(ConfigFile::fo_433_def);               // 352 Defensiv
      else if (ourTactic == TA_Offensive)
        setFormation(ConfigFile::fo_433_off);               // 352 Angriff
      else 
        setFormation(ConfigFile::fo_433);                   // 352
      break;
      
    case FO_532:
      if (debug) printf ("Other formation is FO_532\n");
      if (ourTactic == TA_Defensive)
        setFormation(ConfigFile::fo_433_def);               // 433 Defensiv
      else if (ourTactic == TA_Offensive)
        setFormation(ConfigFile::fo_433_off);               // 433 Angriff
      else 
        setFormation(ConfigFile::fo_433);                   // 433
      break;
      
    case FO_424:
      if (debug) printf ("Other formation is FO_424\n");
      if (ourTactic == TA_Defensive)
        setFormation(ConfigFile::fo_532_def);               // 532 Defensiv
      else if (ourTactic == TA_Offensive)
        setFormation(ConfigFile::fo_532_off);               // 532 Angriff
      else 
        setFormation(ConfigFile::fo_532);                   // 532
      break;
      
    case FO_352:
      if (debug) printf ("Other formation is FO_352\n");
      if (ourTactic == TA_Defensive)
        setFormation(ConfigFile::fo_433_def);               // 433
      else if (ourTactic == TA_Offensive)
        setFormation(ConfigFile::fo_433_off);               // 433
      else 
        setFormation(ConfigFile::fo_433);                   // 433
      break;
      
    case FO_433:
      if (debug) printf ("Other formation is FO_433\n");
      if (ourTactic == TA_Defensive)
        setFormation(ConfigFile::fo_442_def);               // 442 Defensiv
      else if (ourTactic == TA_Offensive)
        setFormation(ConfigFile::fo_442_off);               // 442 Angriff
      else 
        setFormation(ConfigFile::fo_442);                   // 442 GO2004 <Die Raute> German-Open
      break;
      
    case FO_334:
      if (debug) printf ("Other formation is FO_334\n");
      if (ourTactic == TA_Defensive)
        setFormation(ConfigFile::fo_532_def);               // 532 Defensiv
      else if (ourTactic == TA_Offensive)
        setFormation(ConfigFile::fo_532_off);               // 532 Angriff
      else {
        if (fieldside < -1)
          setFormation(ConfigFile::fo_532);
        else 
          setFormation(ConfigFile::fo_442);                 // 442 .. don't panic
      }
      break;
      
    case FO_325:
      if (debug) printf ("Other formation is FO_325\n");
      if (ourTactic == TA_Defensive)
        setFormation(ConfigFile::fo_532_def);               // 532 Defensiv
      else if (ourTactic == TA_Offensive)
        setFormation(ConfigFile::fo_532_off);               // 532 Angriff
      else 
        setFormation(ConfigFile::fo_532);                   // 532
      break;
      
    default:
      if (debug) printf ("Other formation is FO_Unknown\n");
      // handle unknown formations.. 
      if (time >= 100) { // ok... we have somethink detected
        
        if (ourTactic == TA_Defensive) {
          if (subX[2] % 2 == 0) { // even number of attackers
            if (subX[2] == 2) setFormation(ConfigFile::fo_433_def);
            else setFormation(ConfigFile::fo_532_def);
          } else { // odd numbers of attackers
            if (subX[2] == 1) setFormation(ConfigFile::fo_442_def);
            else if (subX[2] == 3) setFormation(ConfigFile::fo_442_def);
            else setFormation(ConfigFile::fo_442_def);
          }
        
        } else if (ourTactic == TA_Offensive) {
          if (subX[2] % 2 == 0) { // even number of attackers
            if (subX[2] == 2) setFormation(ConfigFile::fo_433_off);
            else setFormation(ConfigFile::fo_532_off);
          } else { // odd numbers of attackers
            if (subX[2] == 1) setFormation(ConfigFile::fo_442_off);
            else if (subX[2] == 3) setFormation(ConfigFile::fo_442_off);
            else setFormation(ConfigFile::fo_442_off);
          }
          
        } else {
          if (subX[2] % 2 == 0) { // even number of attackers
            if (subX[2] == 2) setFormation(ConfigFile::fo_433);
            else setFormation(ConfigFile::fo_532);
          } else { // odd numbers of attackers
            if (subX[2] == 1) setFormation(ConfigFile::fo_433);
            else if (subX[2] == 3) setFormation(ConfigFile::fo_442);
            else setFormation(ConfigFile::fo_442_def);
          }
        }
        
      } else
        setFormation(ConfigFile::standardFormation);   
  }
}


void WorldModel::setPlayerTypes (vector<PlayerType> *pt) {
  playerTypes = pt;
}

void WorldModel::getPTMaxValues () {
  if (!(*playerTypes).empty()) {
    vector<PlayerType>::const_iterator pt;
    float ptspeedmax = 0;
    for (pt = (*playerTypes).begin(); pt != (*playerTypes).end(); ++pt){
      ptspeedmax = (pt->getDashPowerRate() * pt->getEffortMax() * 100) / (1 - pt->getPlayerDecay());
      if (ptspeedmax > ptPlayerSpeedMax) {
	ptPlayerSpeedMax = ptspeedmax;
      }
      if (pt->getKickableMargin() > ptKickableMarginMax) {
	ptKickableMarginMax = pt->getKickableMargin();
      }
    }
  }
}

/*
 * Saves the 'worldmodelstate' member-vars of the worldmodel
 */
void WorldModel::saveState(Playmode pm)
{
  WorldModelState *wms;
  wms = static_cast<WorldModelState*>(this);
  wms->playmode = pm;
  wmstates.push_back(*wms);

#if defined (DEBUG_WORLDMODEL)
  DEBUG1 (wmstates.size());
#endif
}


bool WorldModel::balancePlayerTypeDetermination() {
  printf ("Balancing PlayerType Prediction...\n");
  
  // balance count is important to escape this method, if the player types are not balanceable
  // or if it could cost to much time.. see BALANCE_COUNT_MAX
  balanceCount++;
  
  // player types are balanced, if there not more than three players for each type
  bool balanced = true;
  
  // walk through all playertypes
  vector<PlayerType>::const_iterator pt;
  for (pt = (*playerTypes).begin(); pt != (*playerTypes).end(); ++pt) {
    
    // id of this playertype
    int ptId = pt->getTypeId();
    
    // standard type : more than 3 player 
    if( ptId == 0) continue;

    // how many players would get this type
    int typeCount = 0;
    
    // values for each player
    float values[11] = {0};
    
    // then go through the other team
    vector<Player>::iterator p;
    for (p = otherTeam.begin(); p != otherTeam.end(); p++) {
      if (p->isFittestType(ptId)) {
        values[p->getUNum() - 1] = p->getMatchingValue(ptId);
        typeCount++;  
      }
    }
    
    if (typeCount > 3) {
      printf ("  `> Warning: Too many players for type %i found (%i). Requires balancing...\n", ptId, typeCount);
      // the player types are not balanced....
      balanced = false;
      
      // determine which players will get this type...
      float bestValues[3] = {0};
      for (int i = 0; i < 11 ; i++ ) {
        // bestValues[0] has the best value
        // so the min matching value will be stored in bestValues[2]
        if (values[i] > bestValues[0]) {
          bestValues[0] = bestValues[1];
          bestValues[1] = bestValues[2];
          bestValues[2] = values[i];
        } else if (values[i] > bestValues[1]) {
          bestValues[0] = bestValues[1];
          bestValues[1] = values[i];
        } else if (values[i] > bestValues[2]) {
          bestValues[2] = values[i];
        }
      }
      
      // walk again through the players of the other team...
      for (p = otherTeam.begin(); p != otherTeam.end(); p++) {
        // if the matching value of this player is smaller than the min required..
        float pmv = p->getMatchingValue(ptId);
        if (pmv < bestValues[2]) {
          // then change the matching value for this player.. cut 3% of his MatchingValue
          p->setMatchingValue(ptId, pmv * 0.97);
        }
      } 
    }     
  }
  // now being recursive, maybe dangerous ... because its possible that we'll never get balanced types...
  if (!balanced && balanceCount <= BALANCE_COUNT_MAX) balanced = balancePlayerTypeDetermination();
  
  balanceCount = 0;
  return balanced;
}

void WorldModel::printPlayerTypePrediction() {
  vector<Player>::iterator p;
  printf ("\nPlayerType prediction (@cycle %i) for player:\n\n", time);
  int i = 0;
  int lastType = 0;
  for (p = otherTeam.begin(); p != otherTeam.end(); p++) {
    int type = p->getFittestType();
    if (i == 0) lastType = type;
    if (lastType != type)
      printf ("\n");

    printf ("\t%02i  has type:  %i  (matching %.3f)\n", p->getUNum(), type, p->getMatchingValue(type));
    if(i == 0) printf("\n");
    // if(i % 3 == 0) printf("\n");
    lastType = type;
    i++;
  }
  printf ("\n");
}

/*
void WorldModel::checkDuplicatePlayerTypes(){
  return;

  vector<Player>::iterator p;
  int tmp[7] = {0};
  for (p = otherTeam.begin(); p != otherTeam.end(); p++){
    if (p->playerType != -1){
      tmp[p->playerType]++;
    }
  }
  
  bool changed = false;
  for (int i = 0; i < 7; i++){
    if (tmp[i] == 3){
      for (p = otherTeam.begin(); p != otherTeam.end(); p++){
	if (p->playerType == -1){
	  if (p->playerTypePossible[i]){
	    p->playerTypePossible[i] = false;
	    if (p->checkIfDoneWithPlayerTypes())
	      changed = true;
	  }
	}
      }
    }
  }
  if (changed){
    DEBUG1("checkDuplicatePlayerTypes(): duplicate found");
    return checkDuplicatePlayerTypes();
  }
};
*/
